from AccessControl import getSecurityManager
from Acquisition import aq_inner
from Acquisition import aq_parent
from plone.autoform.interfaces import READ_PERMISSIONS_KEY
from plone.dexterity.interfaces import IDexterityContainer
from plone.dexterity.interfaces import IDexterityContent
from plone.dexterity.utils import iterSchemata
from plone.restapi.batching import HypermediaBatch
from plone.restapi.deserializer import boolean_value
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.interfaces import IObjectPrimaryFieldTarget
from plone.restapi.interfaces import IPrimaryFieldTarget
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.serializer.expansion import expandable_elements
from plone.restapi.serializer.nextprev import NextPrevious
from plone.restapi.serializer.working_copy import WorkingCopyInfo
from plone.restapi.services.locking import lock_info
from plone.rfc822.interfaces import IPrimaryFieldInfo
from plone.supermodel.utils import mergedTaggedValueDict
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import base_hasattr
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.interface import Interface
from zope.schema import getFields
from zope.security.interfaces import IPermission


@implementer(ISerializeToJson)
@adapter(IDexterityContent, Interface)
class SerializeToJson:
    def __init__(self, context, request):
        self.context = context
        self.request = request

        self.permission_cache = {}

    def getVersion(self, version):
        if version == "current":
            return self.context
        else:
            repo_tool = getToolByName(self.context, "portal_repository")
            return repo_tool.retrieve(self.context, int(version)).object

    def __call__(self, version=None, include_items=True):
        version = "current" if version is None else version

        obj = self.getVersion(version)
        parent = aq_parent(aq_inner(obj))
        parent_summary = getMultiAdapter(
            (parent, self.request), ISerializeToJsonSummary
        )()
        result = {
            # '@context': 'http://www.w3.org/ns/hydra/context.jsonld',
            "@id": obj.absolute_url(),
            "id": obj.id,
            "@type": obj.portal_type,
            "parent": parent_summary,
            "created": json_compatible(obj.created()),
            "modified": json_compatible(obj.modified()),
            "review_state": self._get_workflow_state(obj),
            "UID": obj.UID(),
            "version": version,
            "layout": self.context.getLayout(),
            "is_folderish": False,
        }

        # Insert next/prev information
        nextprevious = NextPrevious(obj)
        result.update(
            {"previous_item": nextprevious.previous, "next_item": nextprevious.next}
        )

        # Insert working copy information
        baseline, working_copy = WorkingCopyInfo(self.context).get_working_copy_info()
        result.update({"working_copy": working_copy, "working_copy_of": baseline})

        # Insert locking information
        result.update({"lock": lock_info(obj)})

        # Insert expandable elements
        result.update(expandable_elements(self.context, self.request))

        # Insert field values
        for schema in iterSchemata(self.context):
            read_permissions = mergedTaggedValueDict(schema, READ_PERMISSIONS_KEY)

            for name, field in getFields(schema).items():

                if not self.check_permission(read_permissions.get(name), obj):
                    continue

                # serialize the field
                serializer = queryMultiAdapter(
                    (field, obj, self.request), IFieldSerializer
                )
                value = serializer()
                result[json_compatible(name)] = value

        target_url = getMultiAdapter(
            (self.context, self.request), IObjectPrimaryFieldTarget
        )()
        if target_url:
            result["targetUrl"] = target_url

        result["allow_discussion"] = getMultiAdapter(
            (self.context, self.request), name="conversation_view"
        ).enabled()

        return result

    def _get_workflow_state(self, obj):
        wftool = getToolByName(self.context, "portal_workflow")
        review_state = wftool.getInfoFor(ob=obj, name="review_state", default=None)
        return review_state

    def check_permission(self, permission_name, obj):
        if permission_name is None:
            return True

        if permission_name not in self.permission_cache:
            permission = queryUtility(IPermission, name=permission_name)
            if permission is None:
                self.permission_cache[permission_name] = True
            else:
                sm = getSecurityManager()
                self.permission_cache[permission_name] = bool(
                    sm.checkPermission(permission.title, obj)
                )
        return self.permission_cache[permission_name]


@implementer(ISerializeToJson)
@adapter(IDexterityContainer, Interface)
class SerializeFolderToJson(SerializeToJson):
    def _build_query(self):
        path = "/".join(self.context.getPhysicalPath())
        query = {
            "path": {"depth": 1, "query": path},
            "sort_on": "getObjPositionInParent",
        }
        return query

    def __call__(self, version=None, include_items=True):
        folder_metadata = super().__call__(version=version)

        folder_metadata.update({"is_folderish": True})
        result = folder_metadata

        include_items = self.request.form.get("include_items", include_items)
        include_items = boolean_value(include_items)
        if include_items:
            query = self._build_query()

            catalog = getToolByName(self.context, "portal_catalog")
            brains = catalog(query)

            batch = HypermediaBatch(self.request, brains)

            result["items_total"] = batch.items_total
            if batch.links:
                result["batching"] = batch.links

            if "fullobjects" in list(self.request.form):
                result["items"] = getMultiAdapter(
                    (brains, self.request), ISerializeToJson
                )(fullobjects=True)["items"]
            else:
                result["items"] = [
                    getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()
                    for brain in batch
                ]
        return result


@adapter(IDexterityContent, Interface)
@implementer(IObjectPrimaryFieldTarget)
class DexterityObjectPrimaryFieldTarget:
    def __init__(self, context, request):
        self.context = context
        self.request = request

        self.permission_cache = {}

    def __call__(self):
        primary_field_name = self.get_primary_field_name()
        for schema in iterSchemata(self.context):
            read_permissions = mergedTaggedValueDict(schema, READ_PERMISSIONS_KEY)

            for name, field in getFields(schema).items():

                if not self.check_permission(read_permissions.get(name), self.context):
                    continue

                if name != primary_field_name:
                    continue

                target_adapter = queryMultiAdapter(
                    (field, self.context, self.request), IPrimaryFieldTarget
                )
                if target_adapter:
                    target = target_adapter()
                    if target:
                        return target

    def get_primary_field_name(self):
        fieldname = None
        info = None
        try:
            info = IPrimaryFieldInfo(self.context, None)
        except TypeError:
            # No primary field present
            pass
        if info is not None:
            fieldname = info.fieldname
        elif base_hasattr(self.context, "getPrimaryField"):
            field = self.context.getPrimaryField()
            fieldname = field.getName()
        return fieldname

    def check_permission(self, permission_name, obj):
        if permission_name is None:
            return True

        if permission_name not in self.permission_cache:
            permission = queryUtility(IPermission, name=permission_name)
            if permission is None:
                self.permission_cache[permission_name] = True
            else:
                sm = getSecurityManager()
                self.permission_cache[permission_name] = bool(
                    sm.checkPermission(permission.title, obj)
                )
        return self.permission_cache[permission_name]



"""JsonSchema providers."""
from plone.app.textfield.interfaces import IRichText
from plone.restapi.types.interfaces import IJsonSchemaProvider
from plone.restapi.types.utils import get_fieldsets
from plone.restapi.types.utils import get_jsonschema_properties
from plone.restapi.types.utils import get_querysource_url
from plone.restapi.types.utils import get_source_url
from plone.restapi.types.utils import get_vocabulary_url
from plone.restapi.types.utils import get_widget_params
from plone.restapi.types.utils import get_multilingual_directives
from plone.schema import IEmail
from plone.schema import IJSONField
from z3c.formwidget.query.interfaces import IQuerySource
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface
from zope.schema.interfaces import IASCII
from zope.schema.interfaces import IASCIILine
from zope.schema.interfaces import IBool
from zope.schema.interfaces import IBytes
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.interfaces import IDate
from zope.schema.interfaces import IDatetime
from zope.schema.interfaces import IDecimal
from zope.schema.interfaces import IDict
from zope.schema.interfaces import IField
from zope.schema.interfaces import IFloat
from zope.schema.interfaces import IInt
from zope.schema.interfaces import IList
from zope.schema.interfaces import IObject
from zope.schema.interfaces import IPassword
from zope.schema.interfaces import IURI
from zope.schema.interfaces import ISet
from zope.schema.interfaces import IText
from zope.schema.interfaces import ITextLine
from zope.schema.interfaces import ITuple


@adapter(IField, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DefaultJsonSchemaProvider:
    def __init__(self, field, context, request):
        self.field = field.bind(context)
        self.context = context
        self.request = request

    def additional(self):
        """Additional info for field."""
        return {}

    def get_title(self):
        return translate(self.field.title, context=self.request)

    def get_description(self):
        if self.field.description is None:
            return ""

        return translate(self.field.description, context=self.request)

    def get_schema(self):
        """Get jsonschema for field.

        You should override `additional` method to provide more properties
        about the field."""
        schema = {
            "type": self.get_type(),
            "title": self.get_title(),
            "description": self.get_description(),
        }

        widget = self.get_widget()
        if widget:
            schema["widget"] = widget

        factory = self.get_factory()
        if factory:
            schema["factory"] = factory

        widget_options = self.get_widget_params()
        if widget_options:
            schema["widgetOptions"] = widget_options

        multilingual_options = self.get_multilingual_directives()
        if multilingual_options:
            schema["multilingual_options"] = multilingual_options

        if self.field.default is not None:
            schema["default"] = self.field.default

        schema.update(self.additional())
        return schema

    def get_type(self):
        raise NotImplementedError

    def get_factory(self):
        pass

    def get_widget(self):
        return getattr(self.field, "widget", None)

    def get_widget_params(self):
        all_params = get_widget_params([self.field.interface])
        params = all_params.get(self.field.getName(), {})
        if "vocabulary" in params:
            vocab_name = params["vocabulary"]
            params["vocabulary"] = {
                "@id": get_vocabulary_url(vocab_name, self.context, self.request)
            }
        return params

    def get_multilingual_directives(self):
        all_params = get_multilingual_directives([self.field.interface])
        params = all_params.get(self.field.getName(), {})
        return params


@adapter(IBytes, Interface, Interface)
@implementer(IJsonSchemaProvider)
class BytesLineJsonSchemaProvider(DefaultJsonSchemaProvider):
    def get_type(self):
        return "string"

    def get_factory(self):
        return "Text line (String)"


@adapter(ITextLine, Interface, Interface)
@implementer(IJsonSchemaProvider)
class TextLineJsonSchemaProvider(DefaultJsonSchemaProvider):
    def additional(self):
        info = {}
        if self.field.min_length:
            info["minLength"] = self.field.min_length

        if self.field.max_length:
            info["maxLength"] = self.field.max_length

        return info

    def get_type(self):
        return "string"

    def get_factory(self):
        return "Text line (String)"


@adapter(IText, Interface, Interface)
@implementer(IJsonSchemaProvider)
class TextJsonSchemaProvider(TextLineJsonSchemaProvider):
    def get_widget(self):
        return "textarea"

    def get_factory(self):
        return "Text"


@adapter(IEmail, Interface, Interface)
@implementer(IJsonSchemaProvider)
class EmailJsonSchemaProvider(TextLineJsonSchemaProvider):
    def get_widget(self):
        return "email"

    def get_factory(self):
        return "Email"


@adapter(IPassword, Interface, Interface)
@implementer(IJsonSchemaProvider)
class PasswordJsonSchemaProvider(TextLineJsonSchemaProvider):
    def get_widget(self):
        return "password"

    def get_factory(self):
        return "Password"


@adapter(IURI, Interface, Interface)
@implementer(IJsonSchemaProvider)
class URIJsonSchemaProvider(TextLineJsonSchemaProvider):
    def get_widget(self):
        return "url"

    def get_factory(self):
        return "URL"


@adapter(IASCII, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ASCIIJsonSchemaProvider(TextLineJsonSchemaProvider):
    pass


@adapter(IASCIILine, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ASCIILineJsonSchemaProvider(TextLineJsonSchemaProvider):
    pass


@adapter(IFloat, Interface, Interface)
@implementer(IJsonSchemaProvider)
class FloatJsonSchemaProvider(DefaultJsonSchemaProvider):
    def get_type(self):
        return "number"

    def additional(self):
        info = {}
        if self.field.min is not None:
            info["minimum"] = self.field.min

        if self.field.max is not None:
            info["maximum"] = self.field.max

        return info

    def get_factory(self):
        return "Floating-point number"


@adapter(IDecimal, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DecimalJsonSchemaProvider(FloatJsonSchemaProvider):
    pass


@adapter(IInt, Interface, Interface)
@implementer(IJsonSchemaProvider)
class IntegerJsonSchemaProvider(FloatJsonSchemaProvider):
    def get_type(self):
        return "integer"

    def get_factory(self):
        return "Integer"


@adapter(IBool, Interface, Interface)
@implementer(IJsonSchemaProvider)
class BoolJsonSchemaProvider(DefaultJsonSchemaProvider):
    def get_type(self):
        return "boolean"

    def get_factory(self):
        return "Yes/No"


@adapter(ICollection, Interface, Interface)
@implementer(IJsonSchemaProvider)
class CollectionJsonSchemaProvider(DefaultJsonSchemaProvider):
    def get_type(self):
        return "array"

    def get_factory(self):
        map = {
            "RelationList": "Relation List",
            "Set": "Multiple Choice",
            "List": "List",
            "Tuple": "Tuple",
        }

        for key, value in map.items():
            if key in self.field.__repr__():
                return value

        return "Collection"

    def get_items(self):
        """Get items properties."""
        value_type_adapter = getMultiAdapter(
            (self.field.value_type, self.context, self.request), IJsonSchemaProvider
        )

        return value_type_adapter.get_schema()

    def additional(self):
        info = {}
        info["additionalItems"] = True
        if self.field.min_length:
            info["minItems"] = self.field.min_length

        if self.field.max_length:
            info["maxItems"] = self.field.max_length

        info["items"] = self.get_items()

        return info


@adapter(IList, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ListJsonSchemaProvider(CollectionJsonSchemaProvider):
    def additional(self):
        info = super().additional()
        if IChoice.providedBy(self.field.value_type):
            info["uniqueItems"] = True
        else:
            info["uniqueItems"] = False

        return info


@adapter(ISet, Interface, Interface)
@implementer(IJsonSchemaProvider)
class SetJsonSchemaProvider(CollectionJsonSchemaProvider):
    def additional(self):
        info = super().additional()
        info["uniqueItems"] = True
        return info


@adapter(ITuple, Interface, Interface)
@implementer(IJsonSchemaProvider)
class TupleJsonSchemaProvider(SetJsonSchemaProvider):
    pass


@adapter(IChoice, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ChoiceJsonSchemaProvider(DefaultJsonSchemaProvider):

    # optionally prevent rendering all choices in the vocab,
    # ie relatedItems, which contains UUIDs for all content in the site.
    should_render_choices = True

    def get_type(self):
        return "string"

    def get_factory(self):
        map = {"RelationChoice": "Relation Choice", "Choice": "Choice"}

        for key, value in map.items():
            if key in self.field.__repr__():
                return value
        return "Choice"

    def additional(self):
        # Named global vocabulary
        vocab_name = getattr(self.field, "vocabularyName", None)
        if vocab_name:
            return {
                "vocabulary": {
                    "@id": get_vocabulary_url(vocab_name, self.context, self.request)
                }
            }

        # Maybe an unnamed vocabulary or source.
        vocabulary = getattr(self.field, "vocabulary", None)
        if IContextSourceBinder.providedBy(vocabulary):
            vocabulary = vocabulary(self.context)

        # Query source
        if IQuerySource.providedBy(vocabulary):
            return {
                "querysource": {
                    "@id": get_querysource_url(self.field, self.context, self.request)
                }
            }

        # Unamed ISource or vocabulary - render link addressing it via field
        #
        # Even though the URL will point to the @sources endpoint, we also
        # list it under the 'vocabulary' key, because the semantics for an
        # API consumer are exactly the same: A GET to that URL will enumerate
        # terms, and will support batching and filtering by title/token.
        result = {
            "vocabulary": {
                "@id": get_source_url(self.field, self.context, self.request)
            }
        }

        # Optionally inline choices for unnamed sources
        # (this is for BBB, and may eventually be deprecated)
        if hasattr(vocabulary, "__iter__") and self.should_render_choices:
            # choices and enumNames are v5 proposals, for now we implement both
            choices = []
            enum = []
            enum_names = []

            for term in vocabulary:
                if term.title:
                    title = translate(term.title, context=self.request)
                else:
                    title = None
                choices.append((term.token, title))
                enum.append(term.token)
                enum_names.append(title)

            result.update({"enum": enum, "enumNames": enum_names, "choices": choices})

        return result


@adapter(IObject, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ObjectJsonSchemaProvider(DefaultJsonSchemaProvider):

    prefix = ""

    def get_type(self):
        return "object"

    def get_factory(self):
        if self.field.schema.__name__ == "INamedBlobImage":
            return "Image"
        else:
            return "File"

    def get_properties(self):
        if self.prefix:
            prefix = ".".join([self.prefix, self.field.__name__])
        else:
            prefix = self.field.__name__

        context = self.context
        request = self.request
        fieldsets = get_fieldsets(context, request, self.field.schema)
        return get_jsonschema_properties(context, request, fieldsets, prefix)

    def additional(self):
        info = super().additional()
        info["properties"] = self.get_properties()
        return info


@adapter(IDict, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DictJsonSchemaProvider(DefaultJsonSchemaProvider):
    def get_type(self):
        return "dict"

    def additional(self):
        info = {}
        key_type = getMultiAdapter(
            (self.field.key_type, self.context, self.request), IJsonSchemaProvider
        )
        info["key_type"] = {
            "schema": key_type.get_schema(),
            "additional": key_type.additional(),
        }
        value_type = getMultiAdapter(
            (self.field.value_type, self.context, self.request), IJsonSchemaProvider
        )
        info["value_type"] = {
            "schema": value_type.get_schema(),
            "additional": value_type.additional(),
        }
        return info


@adapter(IRichText, Interface, Interface)
@implementer(IJsonSchemaProvider)
class RichTextJsonSchemaProvider(DefaultJsonSchemaProvider):

    prefix = ""

    def get_type(self):
        return "string"

    def get_widget(self):
        return "richtext"

    def get_factory(self):
        return "Rich Text"


@adapter(IDate, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DateJsonSchemaProvider(DefaultJsonSchemaProvider):
    def get_type(self):
        return "string"

    def additional(self):
        info = {}
        if self.field.min is not None:
            info["minimum"] = self.field.min

        if self.field.max is not None:
            info["maximum"] = self.field.max

        return info

    def get_widget(self):
        return "date"

    def get_factory(self):
        return "Date"


@adapter(IDatetime, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DatetimeJsonSchemaProvider(DateJsonSchemaProvider):
    def get_widget(self):
        return "datetime"

    def get_factory(self):
        return "Date/Time"


@adapter(ITuple, Interface, Interface)
@implementer(IJsonSchemaProvider)
class SubjectsFieldJsonSchemaProvider(ChoiceJsonSchemaProvider):
    pass


@adapter(IJSONField, Interface, Interface)
@implementer(IJsonSchemaProvider)
class JSONFieldSchemaProvider(DefaultJsonSchemaProvider):
    def get_type(self):
        return "dict"

    def get_widget(self):
        return getattr(self.field, "widget", False) or "json"

    def get_factory(self):
        return "JSONField"



from plone.restapi.imaging import get_actual_scale
from unittest import TestCase


class TestGetActualScale(TestCase):
    def test_constrains_landscape_to_width(self):
        img = (800, 400)
        bbox = (200, 200)
        self.assertEqual((200, 100), get_actual_scale(img, bbox))

    def test_constrains_portrait_to_height(self):
        img = (400, 800)
        bbox = (200, 200)
        self.assertEqual((100, 200), get_actual_scale(img, bbox))

    def test_doesnt_upscale(self):
        img = (5, 5)
        bbox = (200, 200)
        self.assertEqual((5, 5), get_actual_scale(img, bbox))

    def test_truncates_pixel_dimensions_instead_of_rounding(self):
        img = (215, 56)
        bbox = (64, 64)
        # 64, 16.66
        self.assertEqual((64, 16), get_actual_scale(img, bbox))

    def test_doesnt_produce_zero_pixel_lengths(self):
        img = (1, 1000)
        bbox = (1, 1)
        self.assertEqual((1, 1), get_actual_scale(img, bbox))



from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import TEST_USER_ID
from plone.restapi.behaviors import IBlocks
from plone.restapi.testing import PLONE_RESTAPI_BLOCKS_INTEGRATION_TESTING
from Products.CMFCore.utils import getToolByName

import unittest


class TestProfileBlocks(unittest.TestCase):

    layer = PLONE_RESTAPI_BLOCKS_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        login(self.portal, SITE_OWNER_NAME)

    def test_document_type_has_blocks_behavior_enabled(self):
        self.portal.invokeFactory(
            "Document", id="blocksdoc", title="Document with blocks"
        )
        IBlocks.providedBy(self.portal["blocksdoc"])

    def test_plone_restapi_base_profile_applied(self):
        uf = getToolByName(self.portal, "acl_users")
        self.assertTrue("jwt_auth" in uf)



from plone.app.contentlisting.interfaces import IContentListingObject
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.interfaces import IJSONSummarySerializerMetadata
from plone.restapi.serializer.converters import json_compatible
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
from zope.component import getAllUtilitiesRegisteredFor


@implementer(IJSONSummarySerializerMetadata)
class JSONSummarySerializerMetadata:
    def default_metadata_fields(self):
        return {
            "@id",
            "@type",
            "description",
            "review_state",
            "title",
        }

    def field_accessors(self):
        return {
            "@id": "getURL",
            "@type": "PortalType",
            "description": "Description",
            "title": "Title",
        }

    def non_metadata_attributes(self):
        return {
            "getPath",
            "getURL",
        }

    def blocklisted_attributes(self):
        return {
            "getDataOrigin",
            "getObject",
            "getUserData",
        }


def merge_serializer_metadata_utilities_data():
    """Merge data returned by utilities registered for IJSONSummarySerializerMetadata."""
    serializer_metadata = {
        "default_metadata_fields": set(),
        "field_accessors": {},
        "non_metadata_attributes": set(),
        "blocklisted_attributes": set(),
    }
    utils = getAllUtilitiesRegisteredFor(IJSONSummarySerializerMetadata)
    for name in serializer_metadata.keys():
        for util in utils:
            method = getattr(util, name, None)
            if not method:
                continue
            value = method()
            serializer_metadata[name].update(value)
    return serializer_metadata


@implementer(ISerializeToJsonSummary)
@adapter(Interface, Interface)
class DefaultJSONSummarySerializer:
    """Default ISerializeToJsonSummary adapter.

    Requires context to be adaptable to IContentListingObject, which is
    the case for all content objects providing IContentish.
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request
        # Cache summary_serializer_metadata on request
        metadata = self.request.form.get("summary_serializer_metadata", None)
        if not metadata:
            metadata = merge_serializer_metadata_utilities_data()
            self.request.set("summary_serializer_metadata", metadata)

        self.default_metadata_fields = metadata["default_metadata_fields"]
        self.field_accessors = metadata["field_accessors"]
        self.non_metadata_attributes = metadata["non_metadata_attributes"]
        self.blocklisted_attributes = metadata["blocklisted_attributes"]

    def __call__(self):
        obj = IContentListingObject(self.context)

        summary = {}
        for field in self.metadata_fields():
            if field.startswith("_") or field in self.blocklisted_attributes:
                continue
            accessor = self.field_accessors.get(field, field)
            value = getattr(obj, accessor, None)
            try:
                if callable(value):
                    value = value()
            except WorkflowException:
                summary[field] = None
                continue
            summary[field] = json_compatible(value)
        return summary

    def metadata_fields(self):
        query = self.request.form
        if not query:
            # maybe its a POST request
            query = json_body(self.request)
        additional_metadata_fields = query.get("metadata_fields", [])
        if not isinstance(additional_metadata_fields, list):
            additional_metadata_fields = [additional_metadata_fields]
        additional_metadata_fields = set(additional_metadata_fields)

        if "_all" in additional_metadata_fields:
            fields_cache = self.request.get("_summary_fields_cache", None)
            if fields_cache is None:
                catalog = getToolByName(self.context, "portal_catalog")
                fields_cache = set(catalog.schema()) | self.non_metadata_attributes
                self.request.set("_summary_fields_cache", fields_cache)
            additional_metadata_fields = fields_cache

        return self.default_metadata_fields | additional_metadata_fields


@implementer(ISerializeToJsonSummary)
@adapter(IPloneSiteRoot, Interface)
class SiteRootJSONSummarySerializer:
    """ISerializeToJsonSummary adapter for the Plone Site root."""

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        summary = json_compatible(
            {
                "@id": self.context.absolute_url(),
                "@type": self.context.portal_type,
                "title": self.context.title,
                "description": self.context.description,
            }
        )
        return summary



from plone.outputfilters.browser.resolveuid import uuidToObject
from plone.outputfilters.browser.resolveuid import uuidToURL
from plone.restapi.interfaces import IObjectPrimaryFieldTarget
from zope.component import queryMultiAdapter

import re


RESOLVEUID_RE = re.compile("^[./]*resolve[Uu]id/([^/]*)/?(.*)$")


def uid_to_url(path):
    if not path:
        return ""
    match = RESOLVEUID_RE.match(path)
    if match is None:
        return path

    uid, suffix = match.groups()
    href = uuidToURL(uid)
    if href is None:
        return path
    if suffix:
        href += "/" + suffix
    else:
        target_object = uuidToObject(uid)
        if target_object:
            adapter = queryMultiAdapter(
                (target_object, target_object.REQUEST), IObjectPrimaryFieldTarget
            )
            if adapter and adapter():
                href = adapter()
    return href



from plone import api
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import transaction
import unittest


class TestGroupsEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.gtool = api.portal.get_tool("portal_groups")
        properties = {
            "title": "Plone Team",
            "description": "We are Plone",
            "email": "ploneteam@plone.org",
        }
        self.gtool.addGroup(
            "ploneteam",
            (),
            (),
            properties=properties,
            title=properties["title"],
            description=properties["description"],
        )
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_list_groups(self):
        response = self.api_session.get("/@groups")

        self.assertEqual(200, response.status_code)
        self.assertEqual(5, len(response.json()))
        user_ids = [group["id"] for group in response.json()]
        self.assertIn("Administrators", user_ids)
        self.assertIn("Reviewers", user_ids)
        self.assertIn("AuthenticatedUsers", user_ids)
        self.assertIn("ploneteam", user_ids)
        ptgroup = [x for x in response.json() if x.get("groupname") == "ploneteam"][0]
        self.assertEqual("ploneteam", ptgroup.get("id"))
        self.assertEqual(
            self.portal.absolute_url() + "/@groups/ploneteam", ptgroup.get("@id")
        )
        self.assertEqual("ploneteam@plone.org", ptgroup.get("email"))
        self.assertEqual("Plone Team", ptgroup.get("title"))
        self.assertEqual("We are Plone", ptgroup.get("description"))

        self.assertEqual(ptgroup.get("roles"), ["Authenticated"])

        # We don't want the group members listed in the overview as there
        # might be loads.
        self.assertTrue(
            not any(["users" in group for group in response.json()]),
            "Users key found in groups listing",
        )

    def test_add_group(self):
        response = self.api_session.post(
            "/@groups",
            json={
                "groupname": "fwt",
                "email": "fwt@plone.org",
                "title": "Framework Team",
                "description": "The Plone Framework Team",
                "roles": ["Manager"],
                "groups": ["Administrators"],
                "users": [SITE_OWNER_NAME, TEST_USER_ID],
            },
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        fwt = self.gtool.getGroupById("fwt")
        self.assertEqual("fwt@plone.org", fwt.getProperty("email"))
        self.assertTrue(
            {SITE_OWNER_NAME, TEST_USER_ID}.issubset(set(fwt.getGroupMemberIds())),
            "Userids not found in group",
        )

    def test_add_group_groupname_is_required(self):
        response = self.api_session.post("/@groups", json={"title": "Framework Team"})
        transaction.commit()

        self.assertEqual(400, response.status_code)
        self.assertTrue("\"Property 'groupname' is required" in response.text)

    def test_get_group(self):
        response = self.api_session.get("/@groups/ploneteam")

        self.assertEqual(response.status_code, 200)
        self.assertEqual("ploneteam", response.json().get("id"))
        self.assertEqual(
            self.portal.absolute_url() + "/@groups/ploneteam",
            response.json().get("@id"),
        )
        self.assertEqual("ploneteam@plone.org", response.json().get("email"))
        self.assertEqual("ploneteam@plone.org", response.json().get("email"))
        self.assertEqual("Plone Team", response.json().get("title"))
        self.assertEqual("We are Plone", response.json().get("description"))
        self.assertIn("users", response.json())

    def test_get_search_group_with_filter(self):
        response = self.api_session.get("/@groups", params={"query": "plo"})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        self.assertEqual("ploneteam", response.json()[0].get("id"))
        self.assertEqual(
            self.portal.absolute_url() + "/@groups/ploneteam",
            response.json()[0].get("@id"),
        )
        self.assertEqual("ploneteam@plone.org", response.json()[0].get("email"))

        response = self.api_session.get("/@groups", params={"query": "Auth"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        self.assertEqual("AuthenticatedUsers", response.json()[0].get("id"))

    def test_get_non_existing_group(self):
        response = self.api_session.get("/@groups/non-existing-group")

        self.assertEqual(response.status_code, 404)

    def test_update_group(self):
        ploneteam = self.gtool.getGroupById("ploneteam")
        ploneteam.addMember(SITE_OWNER_NAME)
        transaction.commit()
        self.assertNotIn(TEST_USER_ID, ploneteam.getGroupMemberIds())
        self.assertIn(SITE_OWNER_NAME, ploneteam.getGroupMemberIds())

        payload = {
            "groupname": "ploneteam",
            "email": "ploneteam2@plone.org",
            "users": {TEST_USER_ID: True, SITE_OWNER_NAME: False},
        }
        response = self.api_session.patch("/@groups/ploneteam", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        ploneteam = self.gtool.getGroupById("ploneteam")
        self.assertEqual("ploneteam", ploneteam.id)
        self.assertEqual("Plone Team", ploneteam.getProperty("title"))
        self.assertEqual("ploneteam2@plone.org", ploneteam.getProperty("email"))
        self.assertIn(TEST_USER_ID, ploneteam.getGroupMemberIds())
        self.assertNotIn(SITE_OWNER_NAME, ploneteam.getGroupMemberIds())

    def test_delete_group(self):
        response = self.api_session.delete("/@groups/ploneteam")
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(None, self.gtool.getGroupById("ploneteam"))

    def test_delete_non_existing_group(self):
        response = self.api_session.delete("/@groups/non-existing-group")
        transaction.commit()

        self.assertEqual(response.status_code, 404)



from plone.autoform.interfaces import IFormFieldProvider
from plone.restapi import _
from plone.schema import JSONField
from plone.supermodel import model
from zope.interface import provider

import json


BLOCKS_SCHEMA = json.dumps({"type": "object", "properties": {}})

LAYOUT_SCHEMA = json.dumps(
    {
        "type": "object",
        "properties": {"items": {"type": "array", "items": {"type": "string"}}},
    }
)


@provider(IFormFieldProvider)
class IBlocks(model.Schema):

    model.fieldset("layout", label=_("Layout"), fields=["blocks", "blocks_layout"])

    blocks = JSONField(
        title="Blocks",
        description="The JSON representation of the object blocks information. Must be a JSON object.",  # noqa
        schema=BLOCKS_SCHEMA,
        default={},
        required=False,
    )

    blocks_layout = JSONField(
        title="Blocks Layout",
        description="The JSON representation of the object blocks layout. Must be a JSON array.",  # noqa
        schema=LAYOUT_SCHEMA,
        default={"items": []},
        required=False,
    )



from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from Products.PluggableAuthService.interfaces.plugins import IAuthenticationPlugin


class Logout(Service):
    """Handles logout by invalidating the JWT"""

    def reply(self):
        plugin = None
        acl_users = getToolByName(self, "acl_users")
        mt = getToolByName(self.context, "portal_membership")
        plugins = acl_users._getOb("plugins")
        authenticators = plugins.listPlugins(IAuthenticationPlugin)
        for id_, authenticator in authenticators:
            if authenticator.meta_type == "JWT Authentication Plugin":
                plugin = authenticator
                break

        if plugin is None:
            self.request.response.setStatus(501)
            return dict(
                error=dict(
                    type="Logout failed",
                    message="JWT authentication plugin not installed.",
                )
            )

        if not plugin.store_tokens:
            # Just logout properly from Plone
            mt.logoutUser()
            self.request.response.setStatus(200)
            return super().reply()

        creds = plugin.extractCredentials(self.request)
        if (
            plugin.store_tokens
            and creds
            and "token" in creds
            and plugin.delete_token(creds["token"])
        ):
            # Logout also properly from Plone
            mt.logoutUser()
            self.request.response.setStatus(200)
            return super().reply()

        self.request.response.setStatus(400)
        return dict(error=dict(type="Logout failed", message="Unknown token"))



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.registry.interfaces import IRegistry
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.MailHost.interfaces import IMailHost
from zope.component import getUtility

import transaction
import unittest


class EmailSendEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.mailhost = getUtility(IMailHost)

        registry = getUtility(IRegistry)
        registry["plone.email_from_address"] = "info@plone.org"
        registry["plone.email_from_name"] = "Plone test site"

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        self.anon_api_session = RelativeSession(self.portal_url, test=self)
        self.anon_api_session.headers.update({"Accept": "application/json"})

        transaction.commit()

    def tearDown(self):
        self.api_session.close()
        self.anon_api_session.close()

    def test_email_send(self):
        response = self.api_session.post(
            "/@email-send",
            json={
                "to": "jane@doe.com",
                "from": "john@doe.com",
                "message": "Just want to say hi.",
            },
        )
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        msg = self.mailhost.messages[0]
        if isinstance(msg, bytes) and bytes is not str:
            # Python 3 with Products.MailHost 4.10+
            msg = msg.decode("utf-8")
        self.assertTrue("Subject: =?utf-8?q?A_portal_user_via_Plone_site?=" in msg)
        self.assertTrue("From: info@plone.org" in msg)
        self.assertTrue("Reply-To: john@doe.com" in msg)
        self.assertTrue("Just want to say hi." in msg)

    def test_email_send_all_parameters(self):
        response = self.api_session.post(
            "/@email-send",
            json={
                "to": "jane@doe.com",
                "from": "john@doe.com",
                "message": "Just want to say hi.",
                "name": "John Doe",
                "subject": "This is the subject.",
            },
        )
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        msg = self.mailhost.messages[0]
        if isinstance(msg, bytes) and bytes is not str:
            # Python 3 with Products.MailHost 4.10+
            msg = msg.decode("utf-8")
        self.assertTrue("=?utf-8?q?This_is_the_subject" in msg)
        self.assertTrue("From: info@plone.org" in msg)
        self.assertTrue("John Doe" in msg)
        self.assertTrue("Reply-To: john@doe.com" in msg)
        self.assertTrue("Just want to say hi." in msg)

    def test_email_send_anonymous(self):
        response = self.anon_api_session.post(
            "/@email-send",
            json={
                "to": "jane@doe.com",
                "from": "john@doe.com",
                "message": "Just want to say hi.",
                "name": "John Doe",
                "subject": "This is the subject.",
            },
        )

        self.assertEqual(response.status_code, 401)



from plone.registry.interfaces import IRegistry
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class RegistryGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@registry as parameters
        self.params.append(name)
        return self

    @property
    def _get_record_name(self):
        if len(self.params) != 1:
            raise Exception(
                "Must supply exactly one parameter (dotted name of"
                "the record to be retrieved)"
            )

        return self.params[0]

    def reply(self):
        registry = getUtility(IRegistry)
        if self.params:
            value = registry[self._get_record_name]
            return value
        else:  # batched listing
            serializer = getMultiAdapter((registry, self.request), ISerializeToJson)
            return serializer()



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.utils import createContentInContainer
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from plone.tiles import Tile
from plone.tiles.interfaces import IBasicTile
from plone.tiles.type import TileType
from zope.component import provideAdapter
from zope.component import provideUtility
from zope.component import queryUtility
from zope.interface import Interface

import transaction
import unittest
import zope.schema


class ISampleTile(Interface):
    title = zope.schema.TextLine(title="Title", required=False)


class SampleTile(Tile):

    __name__ = "sample.tile"  # would normally be set by a ZCML handler

    def __call__(self):
        return "<html><body><b>My tile</b></body></html>"


class TestServicesTiles(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        fti = queryUtility(IDexterityFTI, name="Document")
        behavior_list = [a for a in fti.behaviors]
        behavior_list.append("plone.leadimage")
        fti.behaviors = tuple(behavior_list)

        self.doc = createContentInContainer(
            self.portal, "Document", id="doc", title="A document"
        )
        transaction.commit()

        sampleTileType = TileType(
            "sample.tile",
            "Sample tile",
            "cmf.ModifyPortalContent",
            "zope.Public",
            description="A tile used for testing",
            schema=ISampleTile,
            icon="testicon",
        )
        provideUtility(sampleTileType, name="sample.tile")
        provideAdapter(
            SampleTile, (Interface, Interface), IBasicTile, name="sample.tile"
        )

    def tearDown(self):
        self.api_session.close()

    def test_get_available_tiles(self):
        response = self.api_session.get("/@tiles")

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertEqual(len(response), 1)
        self.assertEqual(response[0]["@id"], self.portal_url + "/@tiles/sample.tile")
        self.assertEqual(response[0]["title"], "Sample tile")
        self.assertEqual(response[0]["description"], "A tile used for testing")
        self.assertEqual(response[0]["icon"], "testicon")

    def test_get_tile(self):
        response = self.api_session.get("/@tiles/sample.tile")

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertEqual(response["title"], "Sample tile")
        self.assertEqual(response["properties"]["title"]["title"], "Title")
        self.assertEqual(response["properties"]["title"]["type"], "string")



from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from zope.component import queryMultiAdapter


class ContentGet(Service):
    """Returns a serialized content object."""

    def reply(self):
        serializer = queryMultiAdapter((self.context, self.request), ISerializeToJson)

        if serializer is None:
            self.request.response.setStatus(501)
            return dict(error=dict(message="No serializer available."))

        return serializer(version=self.request.get("version"))



from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

import base64
import requests
import transaction
import unittest


class TestFunctionalAuth(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        login(self.portal, SITE_OWNER_NAME)
        self.private_document = self.portal[
            self.portal.invokeFactory("Document", id="doc1", title="My Document")
        ]
        self.private_document_url = self.private_document.absolute_url()
        transaction.commit()

    def test_login_without_credentials_fails(self):
        response = requests.post(
            self.portal_url + "/@login", headers={"Accept": "application/json"}
        )
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            "Missing credentials", response.json().get("error").get("type")
        )
        self.assertEqual(
            "Login and password must be provided in body.",
            response.json().get("error").get("message"),
        )

    def test_login_with_invalid_credentials_fails(self):
        response = requests.post(
            self.portal_url + "/@login",
            headers={"Accept": "application/json"},
            json={"login": "invalid", "password": "invalid"},
        )
        self.assertEqual(401, response.status_code)
        self.assertEqual(
            "Invalid credentials", response.json().get("error").get("type")
        )
        self.assertEqual(
            "Wrong login and/or password.", response.json().get("error").get("message")
        )

    def test_login_with_valid_credentials_returns_token(self):
        response = requests.post(
            self.portal_url + "/@login",
            headers={"Accept": "application/json"},
            json={"login": TEST_USER_NAME, "password": TEST_USER_PASSWORD},
        )
        self.assertEqual(200, response.status_code)
        self.assertIn("token", response.json())

    def test_api_login_grants_zmi(self):
        """
        Logging in via the API also grants access to the Zope root ZMI.
        """
        session = requests.Session()
        self.addCleanup(session.close)
        login_resp = session.post(
            self.portal_url + "/@login",
            headers={"Accept": "application/json"},
            json={"login": SITE_OWNER_NAME, "password": TEST_USER_PASSWORD},
        )
        self.assertIn(
            "__ac",
            login_resp.cookies,
            "Plone session cookie missing from API login POST response",
        )
        self.assertEqual(
            login_resp.status_code,
            200,
            "Wrong API login response status code",
        )
        self.assertIn(
            "token",
            login_resp.json(),
            "Authentication token missing from API response JSON",
        )

        zmi_resp = session.get(
            self.layer["app"].absolute_url() + "/manage_workspace",
        )
        # Works in the browser when running `$ bin/instance fg` in a `plone.restapi`
        # checkout against `http://localhost:8080/manage_main` but doesn't work in the
        # browser against the test fixture at `http://localhost:55001/manage_main`.  My
        # guess is that there's some subtle difference in the PAS plugin configuration.
        self.skipTest("FIXME: Works in real instance but not test fixture")
        self.assertEqual(
            zmi_resp.status_code,
            200,
            "Wrong ZMI view response status code",
        )
        self.assertTrue(
            '<a href="plone/manage_workspace">' in zmi_resp.text,
            "Wrong ZMI view response content",
        )

    def test_zmi_login_grants_api(self):
        """
        Logging in via the Zope root ZMI also grants access to the API.
        """
        session = requests.Session()
        self.addCleanup(session.close)
        basic_auth_headers = {
            "Authorization": "Basic {}".format(
                base64.b64encode(
                    f"{SITE_OWNER_NAME}:{TEST_USER_PASSWORD}".encode(),
                ).decode()
            )
        }
        zmi_resp = session.get(
            self.layer["app"].absolute_url() + "/manage_workspace",
            headers=basic_auth_headers,
        )
        self.assertEqual(
            zmi_resp.status_code,
            200,
            "Wrong ZMI login response status code",
        )
        self.assertTrue(
            '<a href="plone/manage_workspace">' in zmi_resp.text,
            "Wrong ZMI view response content",
        )

        api_basic_auth_headers = dict(basic_auth_headers)
        api_basic_auth_headers["Accept"] = "application/json"
        api_resp = session.get(
            self.private_document_url,
            headers=api_basic_auth_headers,
        )
        self.assertEqual(
            api_resp.status_code,
            200,
            "Wrong API view response status code",
        )
        api_json = api_resp.json()
        self.assertIn(
            "@id",
            api_json,
            "Plone object id missing from API response JSON",
        )
        self.assertEqual(
            api_json["@id"],
            self.private_document_url,
            "Wrong Plone object URL from API response JSON",
        )

    def test_cookie_login_grants_api(self):
        """
        Logging in via the Plone login form also grants access to the API.
        """
        session = requests.Session()
        self.addCleanup(session.close)
        challenge_resp = session.get(self.private_document_url)
        self.assertEqual(
            challenge_resp.status_code,
            200,
            "Wrong Plone login challenge status code",
        )
        self.assertTrue(
            '<input id="__ac_password" name="__ac_password"' in challenge_resp.text,
            "Plone login challenge response content missing password field",
        )
        login_resp = session.post(
            self.portal_url + "/login",
            data={
                "__ac_name": SITE_OWNER_NAME,
                "__ac_password": TEST_USER_PASSWORD,
                "came_from": "/".join(self.private_document.getPhysicalPath()),
                "buttons.login": "Log in",
            },
        )
        self.assertIn(
            "__ac",
            login_resp.history[0].cookies,
            "Plone session cookie missing form login POST response",
        )
        self.assertEqual(
            login_resp.status_code,
            200,
            "Wrong Plone login response status code",
        )
        self.assertEqual(
            login_resp.url,
            self.private_document_url,
            "Plone login response didn't redirect to original URL",
        )

        api_resp = session.get(
            self.private_document_url,
            headers={"Accept": "application/json"},
        )
        self.assertEqual(
            api_resp.status_code,
            200,
            "Wrong API view response status code",
        )
        api_json = api_resp.json()
        self.assertIn(
            "@id",
            api_json,
            "Plone object id missing from API response JSON",
        )
        self.assertEqual(
            api_json["@id"],
            self.private_document_url,
            "Wrong Plone object URL from API response JSON",
        )

    def test_accessing_private_document_with_valid_token_succeeds(self):
        # login and generate a valid token
        response = requests.post(
            self.portal_url + "/@login",
            headers={"Accept": "application/json"},
            json={"login": TEST_USER_NAME, "password": TEST_USER_PASSWORD},
        )
        valid_token = response.json().get("token")

        # use valid token to access a private resource
        response = requests.get(
            self.private_document_url,
            headers={
                "Accept": "application/json",
                "Authorization": "Bearer " + valid_token,
            },
        )

        self.assertEqual(200, response.status_code)
        self.assertTrue("@id" in response.json())

    def test_accessing_private_document_with_invalid_token_fails(self):
        invalid_token = "abcd1234"
        response = requests.get(
            self.private_document_url,
            headers={
                "Accept": "application/json",
                "Authorization": "Bearer " + invalid_token,
            },
        )

        self.assertEqual(401, response.status_code)
        self.assertEqual("Unauthorized", response.json().get("type"))
        self.assertEqual(
            "You are not authorized to access this resource.",
            response.json().get("message"),
        )

    def test_accessing_private_document_with_expired_token_fails(self):
        # generate an expired token
        self.portal.acl_users.jwt_auth.store_tokens = True
        expired_token = self.portal.acl_users.jwt_auth.create_token(
            "admin", timeout=-60
        )
        transaction.commit()

        response = requests.get(
            self.private_document_url,
            headers={
                "Accept": "application/json",
                "Authorization": "Bearer " + expired_token,
            },
        )

        self.assertEqual(401, response.status_code)
        self.assertEqual("Unauthorized", response.json().get("type"))
        self.assertEqual(
            "You are not authorized to access this resource.",
            response.json().get("message"),
        )



from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.dexterity.interfaces import IDexterityItem
from plone.restapi.exceptions import DeserializationError
from plone.restapi.interfaces import IDeserializeFromJson
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.tests.dxtypes import ITestAnnotationsBehavior
from plone.restapi.tests.mixin_ordering import OrderingMixin
from zExceptions import BadRequest
from zope.component import getMultiAdapter
from zope.component import provideHandler
from zope.lifecycleevent.interfaces import IObjectModifiedEvent

import json
import unittest


class TestDXContentDeserializer(unittest.TestCase, OrderingMixin):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.portal.invokeFactory(
            "DXTestDocument",
            id="doc1",
            test_textline_field="Test Document",
            test_readonly_field="readonly",
        )

        # ordering setup
        self.folder = self.portal[
            self.portal.invokeFactory("Folder", id="folder1", title="Test folder")
        ]

        for x in range(1, 10):
            self.folder.invokeFactory(
                "Document", id="doc" + str(x), title="Test doc " + str(x)
            )

    def deserialize(self, body="{}", validate_all=False, context=None, create=False):
        context = context or self.portal.doc1
        self.request["BODY"] = body
        deserializer = getMultiAdapter((context, self.request), IDeserializeFromJson)
        return deserializer(validate_all=validate_all, create=create)

    def test_deserializer_raises_with_invalid_body(self):
        with self.assertRaises(DeserializationError) as cm:
            self.deserialize(body="Not a JSON object")
        self.assertEqual("No JSON object could be decoded", cm.exception.msg)

    def test_deserializer_raises_with_malformed_body(self):
        with self.assertRaises(DeserializationError) as cm:
            self.deserialize(body="[1,2,3]")
        self.assertEqual("Malformed body", cm.exception.msg)

    def test_deserializer_updates_field_value(self):
        self.deserialize(body='{"test_textline_field": "My Item"}')
        self.assertEqual("My Item", self.portal.doc1.test_textline_field)

    def test_deserializer_ignores_readonly_fields(self):
        self.deserialize(body='{"test_readonly_field": "Foo"}')
        self.assertEqual("readonly", self.portal.doc1.test_readonly_field)

    def test_deserializer_notifies_object_modified(self):
        def handler(obj, event):
            obj._handler_called = True

        provideHandler(handler, (IDexterityItem, IObjectModifiedEvent))
        self.deserialize(body='{"test_textline_field": "My Item"}')
        self.assertTrue(
            getattr(self.portal.doc1, "_handler_called", False),
            "IObjectEditedEvent not notified",
        )

    def test_deserializer_modified_event_contains_descriptions(self):
        def handler(obj, event):
            self.event = event

        provideHandler(handler, (IDexterityItem, IObjectModifiedEvent))
        self.deserialize(body='{"test_textline_field": "My Item"}')
        self.assertEqual(1, len(self.event.descriptions))
        self.assertEqual(
            ("IDXTestDocumentSchema.test_textline_field",),
            self.event.descriptions[0].attributes,
        )

    def test_deserializer_notifies_when_field_is_set_to_null(self):
        def handler(obj, event):
            obj._handler_called = True
            self.event = event

        provideHandler(handler, (IDexterityItem, IObjectModifiedEvent))
        self.deserialize(body='{"test_textline_field": null}')
        self.assertTrue(
            getattr(self.portal.doc1, "_handler_called", False),
            "IObjectModifiedEvent not notified",
        )
        self.assertEqual(
            ("IDXTestDocumentSchema.test_textline_field",),
            self.event.descriptions[0].attributes,
        )

    def test_deserializer_does_not_update_field_without_write_permission(self):
        self.portal.doc1.test_write_permission_field = "Test Write Permission"
        setRoles(self.portal, TEST_USER_ID, ["Member", "Contributor", "Editor"])
        self.deserialize(body='{"test_write_permission_field": "Foo"}')
        self.assertEqual(
            "Test Write Permission", self.portal.doc1.test_write_permission_field
        )

    def test_deserializer_updates_field_with_write_permission(self):
        self.portal.doc1.test_write_permission_field = "Test Write Permission"
        setRoles(self.portal, TEST_USER_ID, ["Member", "Manager"])
        self.deserialize(body='{"test_write_permission_field": "Foo"}')
        self.assertEqual("Foo", self.portal.doc1.test_write_permission_field)

    def test_deserializer_validates_invariant(self):
        with self.assertRaises(BadRequest) as cm:
            self.deserialize(
                body='{"test_invariant_field1": "Foo",'
                ' "test_invariant_field2": "Bar"}'
            )
        self.assertEqual(
            "Must have same values", str(cm.exception.args[0][0]["message"])
        )

    def test_deserializer_updates_behavior_field_value(self):
        self.deserialize(body='{"test_behavior_field": "My Value"}')
        self.assertEqual("My Value", self.portal.doc1.test_behavior_field)

    def test_deserializer_updates_behavior_field_value_in_annotations(self):
        self.deserialize(body='{"test_annotations_behavior_field": "My Value"}')
        self.assertEqual(
            "My Value",
            ITestAnnotationsBehavior(self.portal.doc1).test_annotations_behavior_field,
        )

    def test_deserializer_raises_if_required_value_is_missing(self):
        # Value missing from request
        with self.assertRaises(BadRequest) as cm:
            self.deserialize(
                body='{"test_textline_field": "My Value"}', validate_all=True
            )
        self.assertEqual(
            "Required input is missing.", cm.exception.args[0][0]["message"]
        )

        # An empty string should be considered a missing value
        with self.assertRaises(BadRequest) as cm:
            self.deserialize(body='{"test_textline_field": ""}', validate_all=True)
        self.assertEqual(
            "Required input is missing.", cm.exception.args[0][0]["message"]
        )

    def test_deserializer_succeeds_if_required_value_is_provided(self):
        self.deserialize(body='{"test_required_field": "My Value"}', validate_all=True)
        self.assertEqual("My Value", self.portal.doc1.test_required_field)

    def test_deserializer_does_not_store_default_value(self):
        # XXX: Dexterity has an odd behavior with default values.
        # If a field's value is set to it's default value, it is not stored
        # because z3c.form only updates a field's value if the new value is
        # different from the previous one. Dexterity has a fallback to lookup
        # an attribute's value from the schema's default value if the attribute
        # doesn't exist. Thus the previous value is the default value and the
        # field doesn't get updated if the new value is also the default value.
        # Right now, we want to have the same behavior in the API for
        # consistency reasons.
        self.deserialize(body='{"test_default_value_field": "Default"}')
        self.assertNotIn(
            "test_default_value_field",
            dir(self.portal.doc1),
            "Default value unexpectedly stored.",
        )

    def test_deserializer_passes_validation_with_not_provided_defaults(self):
        self.deserialize(body='{"test_required_field": "My Value"}', validate_all=True)
        self.assertEqual("Default", self.portal.doc1.test_default_value_field)
        self.assertEqual("DefaultFactory", self.portal.doc1.test_default_factory_field)

    def test_deserializer_sets_missing_value_when_receiving_null(self):
        self.deserialize(body='{"test_missing_value_field": null}')
        self.assertEqual("missing", self.portal.doc1.test_missing_value_field)

    def test_deserializer_sets_missing_value_on_required_field(self):
        """We don't set missing_value if the field is required"""
        self.deserialize(body='{"test_missing_value_required_field": "valid value"}')
        with self.assertRaises(BadRequest) as cm:
            self.deserialize(body='{"test_missing_value_required_field": null}')
        self.assertEqual(
            "valid value", self.portal.doc1.test_missing_value_required_field
        )
        self.assertEqual(
            (
                "test_missing_value_required_field is a required field.",
                "Setting it to null is not allowed.",
            ),
            cm.exception.args[0][0]["message"],
        )
        self.assertEqual(
            "test_missing_value_required_field", cm.exception.args[0][0]["field"]
        )

    def test_set_layout(self):
        current_layout = self.portal.doc1.getLayout()
        self.assertNotEqual(current_layout, "my_new_layout")
        self.deserialize(body='{"layout": "my_new_layout"}')
        self.assertEqual("my_new_layout", self.portal.doc1.getLayout())

    def test_deserializer_should_store_default_value_on_content_creation(self):
        # Store default_factory values during content creation. This is specially
        # important for content types like Event, where the default_factory returns
        # a distinct value each time.
        self.portal.invokeFactory(
            "DXTestDocument",
            id="doc_default_value",
        )
        context = self.portal.doc_default_value
        default_value = "DefaultFactory"
        body = {"test_default_factory_field": default_value}
        self.deserialize(body=json.dumps(body), context=context, create=True)
        self.assertIn(
            "test_default_factory_field",
            dir(context),
            "Default value still available.",
        )


class TestDXContentSerializerDeserializer(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.portal.invokeFactory(
            "DXTestDocument",
            id="doc1",
            test_textline_field="Test Document",
            test_readonly_field="readonly",
        )

        self.portal.invokeFactory(
            "DXTestDocument",
            id="doc2",
            test_textline_field="Test Document 2",
            test_readonly_field="readonly",
        )

    def deserialize(self, field, value, validate_all=False, context=None):
        context = context or self.portal.doc1
        body = {}
        body[field] = value
        body = json.dumps(body)
        self.request["BODY"] = body
        deserializer = getMultiAdapter((context, self.request), IDeserializeFromJson)
        return deserializer(validate_all=validate_all)

    def serialize(self, field):
        serializer = getMultiAdapter((self.portal.doc1, self.request), ISerializeToJson)
        return serializer()[field]

    def test_serialize2deserialize_relation(self):
        value = str(self.portal.doc2.UID())
        self.deserialize("test_relationchoice_field", value)

        serialization_value = self.serialize("test_relationchoice_field")

        self.deserialize("test_relationchoice_field", serialization_value)

        self.assertEqual(
            serialization_value["@id"],
            self.portal.doc1.test_relationchoice_field.to_object.absolute_url(),
        )



from plone import api
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.CMFCore import permissions
from Products.CMFCore.ActionInformation import Action
from Products.CMFCore.ActionInformation import ActionCategory

import transaction
import unittest


TEST_CATEGORY_ID = "testcategory"


class TestActions(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def add_category(self, name):
        category = ActionCategory()
        category.id = name
        self.portal_actions._setObject(name, category)
        return category

    def add_action(
        self, category, name, title, icon_expr="", available_expr="", permissions=()
    ):
        action = Action(
            name,
            title=title,
            icon_expr=icon_expr,
            available_expr=available_expr,
            permissions=permissions,
        )
        action.id = name
        category._setObject(name, action)
        return action

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.anon_api_session = RelativeSession(self.portal_url, test=self)
        self.anon_api_session.headers.update({"Accept": "application/json"})

        self.portal_actions = api.portal.get_tool(name="portal_actions")
        existing_ids = self.portal_actions.objectIds()
        self.portal_actions.manage_delObjects(ids=existing_ids)
        self.cat1 = self.add_category("category1")
        self.add_action(
            self.cat1,
            "member_action",
            "Members only",
            available_expr="python:member is not None",
        )
        self.add_action(
            self.cat1,
            "view_action",
            "Action with view permission",
            permissions=(permissions.View,),
        )
        self.add_action(
            self.cat1,
            "manage_action",
            "Action with Manage Portal Content permission",
            permissions=(permissions.ManagePortal,),
        )
        self.cat2 = self.add_category("category2")
        self.cat3 = self.add_category("category3")

        transaction.commit()

    def tearDown(self):
        self.api_session.close()
        self.anon_api_session.close()

    def test_actions_all_categories(self):
        response = self.api_session.get("/@actions")

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertEqual(["category1", "category2", "category3"], sorted(response))

    def test_actions_selected_categories(self):
        response = self.api_session.get(
            "/@actions?categories:list=category1&categories:list=category2"
        )

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertEqual(["category1", "category2"], sorted(response))

    def test_actions_siteroot(self):
        response = self.api_session.get("/@actions")

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertEqual(
            response,
            {
                "category1": [
                    {"title": "Members only", "id": "member_action", "icon": ""},
                    {
                        "title": "Action with view permission",
                        "id": "view_action",
                        "icon": "",
                    },
                    {
                        "title": "Action with Manage Portal Content permission",
                        "id": "manage_action",
                        "icon": "",
                    },
                ],
                "category2": [],
                "category3": [],
            },
        )

    def test_actions_siteroot_anon(self):
        response = self.anon_api_session.get("/@actions")

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertEqual(
            response,
            {
                "category1": [
                    {
                        "title": "Action with view permission",
                        "id": "view_action",
                        "icon": "",
                    }
                ],
                "category2": [],
                "category3": [],
            },
        )

    def test_actions_on_content_object(self):
        self.portal.invokeFactory("Document", id="doc1", title="My Document")
        # we need the category in portal_actions to get additional actions
        # from portal_types
        self.cat_object = self.add_category("object")
        transaction.commit()
        url = "%s/@actions" % self.portal.doc1.absolute_url()
        response = self.api_session.get(url)
        self.assertEqual(response.status_code, 200)
        response = response.json()
        object_action_ids = [action["id"] for action in response["object"]]
        self.assertTrue("view" in object_action_ids)
        self.assertTrue("edit" in object_action_ids)

    def test_actions_on_content_object_anon(self):
        self.portal.invokeFactory("Document", id="doc1", title="My Document")
        api.content.transition(obj=self.portal.doc1, transition="publish")
        # we need the category in portal_actions to get additional actions
        # from portal_types
        self.cat_object = self.add_category("object")
        transaction.commit()
        url = "%s/@actions" % self.portal.doc1.absolute_url()
        response = self.anon_api_session.get(url)
        self.assertEqual(response.status_code, 200)
        response = response.json()
        object_action_ids = [action["id"] for action in response["object"]]
        self.assertTrue("view" in object_action_ids)
        self.assertTrue("edit" not in object_action_ids)



from datetime import date
from datetime import datetime
from datetime import time
from datetime import timedelta
from DateTime import DateTime
from decimal import Decimal
from plone import api
from plone.app.discussion.interfaces import IDiscussionSettings
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.textfield.interfaces import ITransformer
from plone.app.textfield.value import RichTextValue
from plone.namedfile.file import NamedFile
from plone.registry.interfaces import IRegistry
from plone.restapi.interfaces import IExpandableElement
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.tests.test_expansion import ExpandableElementFoo
from plone.uuid.interfaces import IMutableUUID
from Products.CMFCore.utils import getToolByName
from zope.component import getGlobalSiteManager
from zope.component import getMultiAdapter
from zope.component import provideAdapter
from zope.component import queryUtility
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserRequest

import json
import unittest


class AdapterCM:
    """Context manager that will temporarily register an adapter"""

    def __init__(self, adapter, from_, provides):
        self.adapter = adapter
        self.from_ = from_
        self.provides = provides
        self.gsm = getGlobalSiteManager()

    def __enter__(self):
        self.gsm.registerAdapter(self.adapter, self.from_, self.provides)

    def __exit__(self, exc_type, exc_value, traceback):
        self.gsm.unregisterAdapter(self.adapter, self.from_, self.provides)


class TestDXContentSerializer(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        richtext_value = RichTextValue("Käfer", "text/plain", "text/html")

        self.portal.invokeFactory(
            "DXTestDocument",
            id="doc1",
            test_ascii_field="foo",
            test_asciiline_field="foo",
            test_bool_field=True,
            test_bytes_field="\xc3\xa4\xc3\xb6\xc3\xbc",
            test_bytesline_field="\xc3\xa4\xc3\xb6\xc3\xbc",
            test_choice_field="foo",
            test_date_field=date(2015, 7, 15),
            test_datetime_field=datetime(2015, 6, 20, 13, 22, 4),
            test_decimal_field=Decimal(1.1),
            test_dict_field={"foo": "bar", "spam": "eggs", "1": 1},
            test_float_field=1.5,
            test_frozenset_field=frozenset([1, 2, 3]),
            test_int_field=500,
            test_list_field=[1, "two", 3],
            test_set_field={"a", "b", "c"},
            test_text_field="Käfer",
            test_richtext_field=richtext_value,
            test_textline_field="Käfer",
            test_time_field=time(14, 15, 33),
            test_timedelta_field=timedelta(44),
            test_tuple_field=(1, 1),
            test_readonly_field="readonly",
            test_read_permission_field="Secret Stuff",
        )

        self.portal.doc1.creation_date = DateTime("2015-04-27T10:14:48+00:00")
        self.portal.doc1.modification_date = DateTime("2015-04-27T10:24:11+00:00")
        IMutableUUID(self.portal.doc1).set("30314724b77a4ec0abbad03d262837aa")

    def serialize(self, obj=None):
        if obj is None:
            obj = self.portal.doc1
        serializer = getMultiAdapter((obj, self.request), ISerializeToJson)
        return serializer()

    def test_serializer_returns_json_serializeable_object(self):
        obj = self.serialize()
        self.assertTrue(isinstance(json.dumps(obj), str), "Not JSON serializable")

    @unittest.skip("We do not include the context at this point")
    def test_serializer_includes_context(self):
        obj = self.serialize()
        self.assertIn("@context", obj)
        self.assertEqual("http://www.w3.org/ns/hydra/context.jsonld", obj["@context"])

    def test_serializer_includes_json_ld_id(self):
        obj = self.serialize()
        self.assertIn("@id", obj)
        self.assertEqual(self.portal.doc1.absolute_url(), obj["@id"])

    def test_serializer_includes_id(self):
        obj = self.serialize()
        self.assertIn("id", obj)
        self.assertEqual(self.portal.doc1.id, obj["id"])

    def test_serializer_includes_type(self):
        obj = self.serialize()
        self.assertIn("@type", obj)
        self.assertEqual(self.portal.doc1.portal_type, obj["@type"])

    def test_serializer_includes_review_state(self):
        obj = self.serialize()
        self.assertIn("review_state", obj)
        self.assertEqual("private", obj["review_state"])

    def test_serializer_includes_uid(self):
        obj = self.serialize()
        self.assertIn("UID", obj)
        self.assertEqual("30314724b77a4ec0abbad03d262837aa", obj["UID"])

    def test_serializer_includes_creation_date(self):
        obj = self.serialize()
        self.assertIn("created", obj)
        self.assertEqual("2015-04-27T10:14:48+00:00", obj["created"])

    def test_serializer_includes_modification_date(self):
        obj = self.serialize()
        self.assertIn("modified", obj)
        self.assertEqual("2015-04-27T10:24:11+00:00", obj["modified"])

    def test_serializer_ignores_field_without_read_permission(self):
        self.portal.doc1.test_read_permission_field = "Secret Stuff"
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        self.assertNotIn("test_read_permission_field", self.serialize())

    def test_serializer_includes_field_with_read_permission(self):
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        obj = self.serialize()
        self.assertIn("test_read_permission_field", obj)
        self.assertEqual("Secret Stuff", obj["test_read_permission_field"])

    def test_get_layout(self):
        current_layout = self.portal.doc1.getLayout()
        obj = self.serialize()
        self.assertIn("layout", obj)
        self.assertEqual(current_layout, obj["layout"])

    def test_serializer_includes_expansion(self):
        provideAdapter(
            ExpandableElementFoo,
            adapts=(Interface, IBrowserRequest),
            provides=IExpandableElement,
            name="foo",
        )
        obj = self.serialize()
        self.assertIn("foo", obj["@components"])
        self.assertEqual("collapsed", obj["@components"]["foo"])
        gsm = getGlobalSiteManager()
        gsm.unregisterAdapter(
            ExpandableElementFoo,
            (Interface, IBrowserRequest),
            IExpandableElement,
            "foo",
        )

    def test_get_is_folderish(self):
        obj = self.serialize()
        self.assertIn("is_folderish", obj)
        self.assertEqual(False, obj["is_folderish"])

    def test_get_is_folderish_in_folder(self):
        self.portal.invokeFactory("Folder", id="folder")
        serializer = getMultiAdapter(
            (self.portal.folder, self.request), ISerializeToJson
        )
        obj = serializer()
        self.assertIn("is_folderish", obj)
        self.assertEqual(True, obj["is_folderish"])

    def test_nextprev_no_nextprev(self):
        folder = api.content.create(
            container=self.portal,
            type="Folder",
            title="Folder with items",
            description="This is a folder with some documents",
        )
        doc = api.content.create(
            container=folder,
            type="Document",
            title="Item 1",
            description="One item alone in the folder",
        )
        data = self.serialize(doc)
        self.assertEqual({}, data["previous_item"])
        self.assertEqual({}, data["next_item"])

    def test_nextprev_has_prev(self):
        folder = api.content.create(
            container=self.portal,
            type="Folder",
            title="Folder with items",
            description="This is a folder with some documents",
        )
        api.content.create(
            container=folder,
            type="Document",
            title="Item 1",
            description="Previous item",
        )
        doc = api.content.create(
            container=folder,
            type="Document",
            title="Item 2",
            description="Current item",
        )
        data = self.serialize(doc)
        self.assertEqual(
            {
                "@id": "http://nohost/plone/folder-with-items/item-1",
                "@type": "Document",
                "title": "Item 1",
                "description": "Previous item",
            },
            data["previous_item"],
        )
        self.assertEqual({}, data["next_item"])

    def test_nextprev_has_next(self):
        folder = api.content.create(
            container=self.portal,
            type="Folder",
            title="Folder with items",
            description="This is a folder with some documents",
        )
        doc = api.content.create(
            container=folder,
            type="Document",
            title="Item 1",
            description="Current item",
        )
        api.content.create(
            container=folder, type="Document", title="Item 2", description="Next item"
        )
        data = self.serialize(doc)
        self.assertEqual({}, data["previous_item"])
        self.assertEqual(
            {
                "@id": "http://nohost/plone/folder-with-items/item-2",
                "@type": "Document",
                "title": "Item 2",
                "description": "Next item",
            },
            data["next_item"],
        )

    def test_nextprev_has_nextprev(self):
        folder = api.content.create(
            container=self.portal,
            type="Folder",
            title="Folder with items",
            description="This is a folder with some documents",
        )
        api.content.create(
            container=folder,
            type="Document",
            title="Item 1",
            description="Previous item",
        )
        doc = api.content.create(
            container=folder,
            type="Document",
            title="Item 2",
            description="Current item",
        )
        api.content.create(
            container=folder, type="Document", title="Item 3", description="Next item"
        )
        data = self.serialize(doc)
        self.assertEqual(
            {
                "@id": "http://nohost/plone/folder-with-items/item-1",
                "@type": "Document",
                "title": "Item 1",
                "description": "Previous item",
            },
            data["previous_item"],
        )
        self.assertEqual(
            {
                "@id": "http://nohost/plone/folder-with-items/item-3",
                "@type": "Document",
                "title": "Item 3",
                "description": "Next item",
            },
            data["next_item"],
        )

    def test_nextprev_root_no_nextprev(self):
        data = self.serialize()
        self.assertEqual({}, data["previous_item"])
        self.assertEqual({}, data["next_item"])

    def test_nextprev_root_has_prev(self):
        doc = api.content.create(
            container=self.portal,
            type="Document",
            title="Item 2",
            description="Current item",
        )
        data = self.serialize(doc)
        self.assertEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "title": "",
                "description": "",
            },
            data["previous_item"],
        )
        self.assertEqual({}, data["next_item"])

    def test_nextprev_root_has_next(self):
        api.content.create(
            container=self.portal,
            type="Document",
            title="Item 2",
            description="Next item",
        )
        data = self.serialize()
        self.assertEqual({}, data["previous_item"])
        self.assertEqual(
            {
                "@id": "http://nohost/plone/item-2",
                "@type": "Document",
                "title": "Item 2",
                "description": "Next item",
            },
            data["next_item"],
        )

    def test_nextprev_root_has_nextprev(self):
        api.content.create(
            container=self.portal,
            type="Document",
            title="Item 1",
            description="Previous item",
        )
        doc = api.content.create(
            container=self.portal,
            type="Document",
            title="Item 2",
            description="Current item",
        )
        api.content.create(
            container=self.portal,
            type="Document",
            title="Item 3",
            description="Next item",
        )
        data = self.serialize(doc)
        self.assertEqual(
            {
                "@id": "http://nohost/plone/item-1",
                "@type": "Document",
                "title": "Item 1",
                "description": "Previous item",
            },
            data["previous_item"],
        )
        self.assertEqual(
            {
                "@id": "http://nohost/plone/item-3",
                "@type": "Document",
                "title": "Item 3",
                "description": "Next item",
            },
            data["next_item"],
        )

    def test_nextprev_unordered_folder(self):
        folder = api.content.create(
            container=self.portal,
            type="Folder",
            title="Folder with items",
            description="This is a folder with some documents",
        )
        folder.setOrdering("unordered")
        doc = api.content.create(
            container=folder,
            type="Document",
            title="Item 1",
            description="One item alone in the folder",
        )
        data = self.serialize(doc)
        self.assertEqual({}, data["previous_item"])
        self.assertEqual({}, data["next_item"])

    def test_richtext_serializer_context(self):
        """This checks if the context is passed in correctly.

        We define a ITransformer, which returns the contexts portal_type.
        This is then verfied.
        """

        class RichtextTransform:
            """RichttextValue to show that the context is correctly passed
            in throughout the stack.
            """

            def __init__(self, context):
                self.context = context

            def __call__(self, value, mime_type):
                return self.context.portal_type

        with AdapterCM(RichtextTransform, (Interface,), ITransformer):
            obj = self.serialize()

        self.assertEqual(
            obj["test_richtext_field"]["data"], self.portal.doc1.portal_type
        )

    def test_allow_discussion_by_default(self):
        """Not globally addable, not fti enabled, not obj instance enabled"""
        self.portal.invokeFactory("Document", id="doc2")
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(False, obj["allow_discussion"])

    def test_allow_discussion_obj_instance_allows_but_not_global_enabled(self):
        self.portal.invokeFactory("Document", id="doc2")
        self.portal.doc2.allow_discussion = True
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(False, obj["allow_discussion"])

    def test_allow_discussion_fti_allows_not_global_enabled(self):
        self.portal.invokeFactory("Document", id="doc2")
        portal_types = getToolByName(self.portal, "portal_types")
        document_fti = getattr(portal_types, self.portal.doc2.portal_type)
        document_fti.allow_discussion = True
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(False, obj["allow_discussion"])

    def test_allow_discussion_allows_global_enabled_but_nothing_else(self):
        self.portal.invokeFactory("Document", id="doc2")
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(False, obj["allow_discussion"])

    def test_allow_discussion_obj_instance_allows_global_enabled(self):
        self.portal.invokeFactory("Document", id="doc2")
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        self.portal.doc2.allow_discussion = True
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(True, obj["allow_discussion"])

    def test_allow_discussion_obj_instance_not_set_global_enabled(self):
        self.portal.invokeFactory("Document", id="doc2")
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(False, obj["allow_discussion"])

    def test_allow_discussion_fti_allows_allows_global_enabled(self):
        self.portal.invokeFactory("Document", id="doc2")
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        portal_types = getToolByName(self.portal, "portal_types")
        document_fti = getattr(portal_types, self.portal.doc2.portal_type)
        document_fti.allow_discussion = True
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(True, obj["allow_discussion"])

    def test_allow_discussion_fti_allows_allows_global_enabled_but_no_instance_allowed(
        self,
    ):  # noqa
        self.portal.invokeFactory("Document", id="doc2")
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        portal_types = getToolByName(self.portal, "portal_types")
        document_fti = getattr(portal_types, self.portal.doc2.portal_type)
        document_fti.allow_discussion = True
        self.portal.doc2.allow_discussion = False

        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(False, obj["allow_discussion"])

    def test_allow_discussion_fti_allows_allows_global_enabled_but_no_instance_set(
        self,
    ):  # noqa
        self.portal.invokeFactory("Document", id="doc2")
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        portal_types = getToolByName(self.portal, "portal_types")
        document_fti = getattr(portal_types, self.portal.doc2.portal_type)
        document_fti.allow_discussion = True

        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(True, obj["allow_discussion"])

    def test_allow_discussion_fti_disallows_allows_global_enabled_but_instance_allowed(
        self,
    ):  # noqa
        self.portal.invokeFactory("Document", id="doc2")
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        portal_types = getToolByName(self.portal, "portal_types")
        document_fti = getattr(portal_types, self.portal.doc2.portal_type)
        document_fti.allow_discussion = False
        self.portal.doc2.allow_discussion = True

        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        obj = serializer()

        self.assertIn("allow_discussion", obj)
        self.assertEqual(True, obj["allow_discussion"])

    def test_allow_discussion_global_enabled_but_instance_has_no_discussion_behavior(
        self,
    ):  # noqa
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True

        obj = self.serialize()
        self.assertIn("allow_discussion", obj)
        self.assertEqual(False, obj["allow_discussion"])


class TestDXContentPrimaryFieldTargetUrl(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.portal.invokeFactory(
            "DXTestDocument",
            id="doc1",
            test_primary_namedfile_field=NamedFile(
                data="Spam and eggs", contentType="text/plain", filename="test.txt"
            ),
        )

        self.portal.invokeFactory(
            "DXTestDocument",
            id="doc2",
            test_primary_namedfile_field=None,
        )

        self.portal.doc1.creation_date = DateTime("2015-04-27T10:14:48+00:00")
        self.portal.doc1.modification_date = DateTime("2015-04-27T10:24:11+00:00")
        IMutableUUID(self.portal.doc1).set("30314724b77a4ec0abbad03d262837aa")

    def serialize(self):
        serializer = getMultiAdapter((self.portal.doc1, self.request), ISerializeToJson)
        return serializer()

    def test_primary_field_target(self):
        logout()
        serializer = getMultiAdapter((self.portal.doc1, self.request), ISerializeToJson)
        data = serializer()
        self.assertIn("targetUrl", data)
        download_url = "/".join(
            [
                self.portal.doc1.absolute_url(),
                "@@download/test_primary_namedfile_field",
            ]
        )
        self.assertEqual(data["targetUrl"], download_url)

    def test_primary_field_target_without_file(self):
        logout()
        serializer = getMultiAdapter((self.portal.doc2, self.request), ISerializeToJson)
        data = serializer()
        self.assertNotIn("targetUrl", data)

    def test_primary_field_target_with_edit_permissions(self):
        serializer = getMultiAdapter((self.portal.doc1, self.request), ISerializeToJson)
        data = serializer()
        self.assertNotIn("targetUrl", data)



from collections import OrderedDict
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import IPloneRestapiLayer
from plone.restapi.services import Service
from plone.restapi.types.utils import add_field
from plone.restapi.types.utils import add_fieldset
from plone.restapi.types.utils import delete_field
from plone.restapi.types.utils import delete_fieldset
from plone.restapi.types.utils import get_info_for_type
from plone.restapi.types.utils import serializeSchema
from plone.restapi.types.utils import update_field
from plone.restapi.types.utils import update_fieldset
from plone.supermodel.interfaces import FIELDSETS_KEY
from Products.CMFCore.utils import getToolByName
from zExceptions import BadRequest
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import noLongerProvides
from zope.publisher.interfaces import IPublishTraverse
import plone.protect.interfaces


@implementer(IPublishTraverse)
class TypesPut(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Treat any path segments after /@types as parameters
        self.params.append(name)
        return self

    def update_layouts(self, name, data):
        layouts = data.get("layouts", [])
        if not layouts:
            return

        ttool = getToolByName(self.context, "portal_types")
        ttool[name].view_methods = tuple(layouts)

    def remove_fieldsets(self, ctype, data):
        fieldsets = [f.get("id") for f in data.get("fieldsets", [])]
        existing = {
            f.__name__ for f in ctype.schema.queryTaggedValue(FIELDSETS_KEY, [])
        }

        for fieldset in existing:
            if fieldset not in fieldsets:
                delete_fieldset(ctype, self.request, fieldset)

    def add_fieldsets(self, ctype, data):
        fieldsets = OrderedDict((f.get("id"), f) for f in data.get("fieldsets", []))
        info = get_info_for_type(ctype, self.request, ctype.getId())
        existing = {f.get("id") for f in info.get("fieldsets", [])}
        for name, fieldset in fieldsets.items():
            if name not in existing:
                add_fieldset(ctype, self.request, fieldset)

    def remove_fields(self, ctype, data):
        fields = data.get("properties", {})
        existing = {name for name in ctype.schema}
        for name in existing:
            if name not in fields:
                delete_field(ctype, self.request, name)

    def add_fields(self, ctype, data):
        allow = [ctype.schema.__identifier__, ""]
        required = data.get("required", [])
        for name, field in data.get("properties", {}).items():
            if name in ctype.schema:
                continue

            behavior = field.get("behavior", "")
            if behavior not in allow:
                continue

            if name in required:
                field["required"] = True
            field["id"] = name
            add_field(ctype, self.request, field)

    def update_fieldsets(self, ctype, data):
        existing = {
            f.__name__ for f in ctype.schema.queryTaggedValue(FIELDSETS_KEY, [])
        }
        existing.add("default")
        for fieldset in data.get("fieldsets", []):
            if fieldset.get("id") not in existing:
                continue
            update_fieldset(ctype, self.request, fieldset)

    def update_fields(self, ctype, data):
        for name, field in data.get("properties", {}).items():
            if name not in ctype.schema:
                continue

            field["id"] = name
            update_field(ctype, self.request, field)

    def reply(self):
        if not self.params:
            raise BadRequest("Missing parameter typename")

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        # Make sure we get the right dexterity-types adapter
        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        data = json_body(self.request)
        name = self.params.pop()
        dtypes = queryMultiAdapter((self.context, self.request), name="dexterity-types")
        ctype = dtypes.publishTraverse(self.request, name)

        self.update_layouts(name, data)
        self.remove_fields(ctype, data)
        self.remove_fieldsets(ctype, data)
        self.add_fields(ctype, data)
        self.add_fieldsets(ctype, data)
        self.update_fields(ctype, data)
        self.update_fieldsets(ctype, data)

        serializeSchema(ctype.schema)
        return self.reply_no_content()



from plone.dexterity.interfaces import IDexterityContent
from plone.registry.interfaces import IRegistry
from plone.restapi.controlpanels import IControlpanel
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.types import utils
from zope.component import adapter
from zope.component import getUtility
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import noLongerProvides

import zope.schema


SERVICE_ID = "@controlpanels"


@implementer(ISerializeToJsonSummary)
@adapter(IControlpanel)
class ControlpanelSummarySerializeToJson:
    def __init__(self, controlpanel):
        self.controlpanel = controlpanel

    def __call__(self):
        return {
            "@id": "{}/{}/{}".format(
                self.controlpanel.context.absolute_url(),
                SERVICE_ID,
                self.controlpanel.__name__,
            ),
            "title": self.controlpanel.title,
            "group": self.controlpanel.group,
        }


def get_jsonschema_for_controlpanel(controlpanel, context, request, form=None):
    """Build a complete JSON schema for the given controlpanel."""
    if not form:
        schema = controlpanel.schema
        fieldsets = utils.get_fieldsets(context, request, schema)
    else:
        fieldsets = utils.get_form_fieldsets(form)

    # Build JSON schema properties
    properties = utils.get_jsonschema_properties(context, request, fieldsets)

    # Determine required fields
    required = []
    for field in utils.iter_fields(fieldsets):
        if field.field.required:
            required.append(field.field.getName())

    # Include field modes
    for field in utils.iter_fields(fieldsets):
        if field.mode:
            properties[field.field.getName()]["mode"] = field.mode

    return {
        "type": "object",
        "properties": properties,
        "required": required,
        "fieldsets": utils.get_fieldset_infos(fieldsets),
    }


@implementer(ISerializeToJson)
@adapter(IControlpanel)
class ControlpanelSerializeToJson:
    def __init__(self, controlpanel):
        self.controlpanel = controlpanel
        self.schema = self.controlpanel.schema
        self.schema_prefix = self.controlpanel.schema_prefix

        self.registry = getUtility(IRegistry)

    def __call__(self):
        json_schema = get_jsonschema_for_controlpanel(
            self.controlpanel, self.controlpanel.context, self.controlpanel.request
        )

        proxy = self.registry.forInterface(self.schema, prefix=self.schema_prefix)

        # Temporarily provide IDexterityContent, so we can use DX field
        # serializers
        alsoProvides(proxy, IDexterityContent)

        json_data = {}
        for name, field in zope.schema.getFields(self.schema).items():
            serializer = queryMultiAdapter(
                (field, proxy, self.controlpanel.request), IFieldSerializer
            )
            if serializer:
                value = serializer()
            else:
                value = getattr(proxy, name, None)
            json_data[json_compatible(name)] = value

        noLongerProvides(proxy, IDexterityContent)

        # JSON schema
        return {
            "@id": "{}/{}/{}".format(
                self.controlpanel.context.absolute_url(),
                SERVICE_ID,
                self.controlpanel.__name__,
            ),
            "title": self.controlpanel.title,
            "group": self.controlpanel.group,
            "schema": json_schema,
            "data": json_data,
        }



from base64 import b64encode
from DateTime import DateTime
from io import BytesIO
from OFS.interfaces import IObjectWillBeAddedEvent
from plone import api
from plone.app.testing import login
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.rest.cors import CORSPolicy
from plone.rest.interfaces import ICORSPolicy
from plone.restapi.services.content.tus import TUSUpload
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import getGlobalSiteManager
from zope.component import provideAdapter
from zope.interface import Interface
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectCreatedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.publisher.interfaces.browser import IBrowserRequest

import os
import shutil
import tempfile
import transaction
import unittest


UPLOAD_DATA = b"abcdefgh"
UPLOAD_MIMETYPE = "text/plain"
UPLOAD_FILENAME = "test.txt"
UPLOAD_LENGTH = len(UPLOAD_DATA)

UPLOAD_PDF_MIMETYPE = "application/pdf"
UPLOAD_PDF_FILENAME = "file.pdf"


def _base64_str(s):
    if not isinstance(s, bytes):
        s = s.encode("utf-8")
    s = b64encode(s)
    if not isinstance(s, str):
        s = s.decode("utf-8")
    return s


def _prepare_metadata(filename, content_type):
    return "filename {},content-type {}".format(
        _base64_str(filename), _base64_str(content_type)
    )


class TestTUS(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        login(self.portal, SITE_OWNER_NAME)

        self.folder = api.content.create(
            container=self.portal, type="Folder", id="testfolder", title="Testfolder"
        )
        self.upload_url = f"{self.folder.absolute_url()}/@tus-upload"
        transaction.commit()

        self.api_session = RelativeSession(self.portal.absolute_url(), test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def get_tus_uid_from_url(self, url):
        uid = url.rsplit("/", 1)[-1]
        assert len(uid) == 32
        return uid

    def get_tus_upload_instance(self, uid):
        return TUSUpload(uid)

    def test_tus_option_headers(self):
        response = self.api_session.options(self.upload_url)
        headers = response.headers
        self.assertEqual(response.status_code, 204)
        self.assertEqual(headers["Tus-Version"], "1.0.0")
        self.assertEqual(headers["Tus-Extension"], "creation,expiration")
        self.assertEqual(headers["Tus-Resumable"], "1.0.0")

    def test_tus_post_without_version_header_returns_412(self):
        response = self.api_session.post(self.upload_url)
        self.assertEqual(412, response.status_code)

    def test_tus_post_with_wrong_version_header_returns_412(self):
        response = self.api_session.post(
            self.upload_url, headers={"Tus-Resumable": "0.2.2"}
        )
        self.assertEqual(412, response.status_code)

    def test_tus_post_initialization_requires_header_length(self):
        response = self.api_session.post(
            self.upload_url, headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertEqual(response.json()["error"]["type"], "Bad Request")
        self.assertEqual(
            response.json()["error"]["message"],
            "Missing or invalid Upload-Length header",
        )
        self.assertEqual(response.status_code, 400)

    def test_tus_post_initialization(self):
        response = self.api_session.post(
            self.upload_url,
            headers={"Tus-Resumable": "1.0.0", "Upload-Length": str(UPLOAD_LENGTH)},
        )
        self.assertEqual(response.status_code, 201)
        location = response.headers["Location"]
        url_base, uid = location.rsplit("/", 1)
        self.assertEqual(url_base, self.upload_url)
        self.assertEqual(len(uid), 32)
        upload = TUSUpload(uid)
        stored_metadata = upload.metadata()
        self.assertEqual(stored_metadata, {"length": 8, "mode": "create"})
        upload.cleanup()

    def test_tus_post_initialization_with_metadata(self):
        metadata = _prepare_metadata(UPLOAD_FILENAME, UPLOAD_MIMETYPE)
        response = self.api_session.post(
            self.upload_url,
            headers={
                "Tus-Resumable": "1.0.0",
                "Upload-Length": str(UPLOAD_LENGTH),
                "Upload-Metadata": metadata,
            },
        )
        self.assertEqual(response.status_code, 201)
        uid = self.get_tus_uid_from_url(response.headers["Location"])
        upload = TUSUpload(uid)
        stored_metadata = upload.metadata()
        self.assertEqual(
            stored_metadata,
            {
                "content-type": "text/plain",
                "filename": "test.txt",
                "length": 8,
                "mode": "create",
            },
        )
        upload.cleanup()

    def test_tus_post_replace(self):
        self.file = api.content.create(
            container=self.portal, type="File", id="testfile", title="Testfile"
        )
        transaction.commit()
        response = self.api_session.post(
            f"{self.file.absolute_url()}/@tus-replace",
            headers={"Tus-Resumable": "1.0.0", "Upload-Length": str(UPLOAD_LENGTH)},
        )
        self.assertEqual(response.status_code, 201)
        location = response.headers["Location"]
        url_base, uid = location.rsplit("/", 1)
        upload = TUSUpload(uid)
        stored_metadata = upload.metadata()
        self.assertEqual(stored_metadata, {"length": 8, "mode": "replace"})
        upload.cleanup()

    def test_tus_head_on_not_existing_resource_returns_404(self):
        response = self.api_session.head(
            self.upload_url + "/myuid/123", headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertEqual(404, response.status_code)
        response = self.api_session.head(
            self.upload_url + "/non-existing-uid", headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertEqual(404, response.status_code)
        response = self.api_session.head(
            self.upload_url, headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertEqual(404, response.status_code)

    def test_tus_head_with_unsupported_version_returns_412(self):
        tus = TUSUpload("myuid", {"length": 2048})
        response = self.api_session.head(
            self.upload_url + "/myuid", headers={"Tus-Resumable": "0.2.2"}
        )
        self.assertEqual(412, response.status_code)
        tus.cleanup()

    def test_tus_head_response_includes_required_headers(self):
        tus = TUSUpload("myuid", {"length": 2048})
        response = self.api_session.head(
            self.upload_url + "/myuid", headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertIn("Upload-Length", response.headers)
        self.assertEqual("2048", response.headers["Upload-Length"])
        self.assertIn("Upload-Offset", response.headers)
        self.assertIn("Tus-Resumable", response.headers)
        self.assertIn("Cache-Control", response.headers)
        tus.cleanup()

    def test_head_in_create_mode_without_add_permission_raises_401(self):
        self.folder.manage_permission("Add portal content", [], 0)
        transaction.commit()
        tus = TUSUpload("myuid", {"mode": "create", "length": 12})
        response = self.api_session.head(
            self.upload_url + "/myuid",
            headers={"Tus-Resumable": "1.0.0", "Upload-Offset": "0"},
        )
        self.assertEqual(401, response.status_code)
        tus.cleanup()

    def test_head_in_replace_mode_without_modify_permission_raises_401(self):
        self.folder.manage_permission("Modify portal content", [], 0)
        transaction.commit()
        tus = TUSUpload("myuid", {"mode": "replace", "length": 12})
        response = self.api_session.head(
            self.upload_url + "/myuid",
            headers={"Tus-Resumable": "1.0.0", "Upload-Offset": "0"},
        )
        self.assertEqual(401, response.status_code)
        tus.cleanup()

    def test_tus_patch_on_not_existing_resource_returns_404(self):
        response = self.api_session.patch(
            self.upload_url + "/myuid/123", headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertEqual(404, response.status_code)
        response = self.api_session.patch(
            self.upload_url + "/myuid", headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertEqual(404, response.status_code)
        response = self.api_session.patch(
            self.upload_url, headers={"Tus-Resumable": "1.0.0"}
        )
        self.assertEqual(404, response.status_code)

    def test_tus_patch_with_unsupported_version_returns_412(self):
        tus = TUSUpload("myuid", {"length": 2048})
        response = self.api_session.patch(
            self.upload_url + "/myuid", headers={"Tus-Resumable": "0.2.2"}
        )
        self.assertEqual(412, response.status_code)
        tus.cleanup()

    def test_tus_patch_with_unsupported_content_type_returns_400(self):
        tus = TUSUpload("myuid", {"length": 2048})
        response = self.api_session.patch(
            self.upload_url + "/myuid",
            headers={"Tus-Resumable": "1.0.0", "Content-Type": "application/json"},
        )
        self.assertEqual(400, response.status_code)
        tus.cleanup()

    def test_tus_patch_with_invalid_offset_returns_400(self):
        tus = TUSUpload("myuid", {"length": 2048})
        response = self.api_session.patch(
            self.upload_url + "/myuid",
            headers={
                "Tus-Resumable": "1.0.0",
                "Content-Type": "application/offset+octet-stream",
            },
        )
        self.assertEqual(400, response.status_code)
        tus.cleanup()

    def test_tus_patch_unfinished_upload_returns_expires_header(self):
        tus = TUSUpload("myuid", {"length": 2048})
        response = self.api_session.patch(
            self.upload_url + "/myuid",
            headers={
                "Tus-Resumable": "1.0.0",
                "Content-Type": "application/offset+octet-stream",
                "Upload-Offset": "0",
            },
            data=BytesIO(b"abcdefghijkl"),
        )
        self.assertEqual(204, response.status_code)
        self.assertIn("Upload-Expires", response.headers)
        tus.cleanup()

    def test_tus_patch_non_primary_field(self):
        tus = TUSUpload(
            "myuid",
            {
                "@type": "DXTestDocument",
                "length": 12,
                "fieldname": "test_namedblobfile_field",
            },
        )
        response = self.api_session.patch(
            self.upload_url + "/myuid",
            headers={
                "Tus-Resumable": "1.0.0",
                "Content-Type": "application/offset+octet-stream",
                "Upload-Offset": "0",
            },
            data=BytesIO(b"abcdefghijkl"),
        )

        self.assertEqual(204, response.status_code)
        transaction.commit()
        self.assertEqual(1, len(self.folder.objectIds()))
        id_ = self.folder.objectIds()[0]
        self.assertEqual(
            b"abcdefghijkl", self.folder[id_].test_namedblobfile_field.data
        )
        tus.cleanup()

    def test_patch_in_create_mode_without_add_permission_raises_401(self):
        self.folder.manage_permission("Add portal content", [], 0)
        transaction.commit()
        tus = TUSUpload("myuid", {"mode": "create", "length": 12})
        response = self.api_session.patch(
            self.upload_url + "/myuid",
            headers={
                "Tus-Resumable": "1.0.0",
                "Content-Type": "application/offset+octet-stream",
                "Upload-Offset": "0",
            },
            data=BytesIO(b"abcdefghijkl"),
        )
        self.assertEqual(401, response.status_code)
        tus.cleanup()

    def test_patch_in_replace_mode_without_modify_permission_raises_401(self):
        self.folder.manage_permission("Modify portal content", [], 0)
        transaction.commit()
        tus = TUSUpload("myuid", {"mode": "replace", "length": 12})
        response = self.api_session.patch(
            self.upload_url + "/myuid",
            headers={
                "Tus-Resumable": "1.0.0",
                "Content-Type": "application/offset+octet-stream",
                "Upload-Offset": "0",
            },
            data=BytesIO(b"abcdefghijkl"),
        )
        self.assertEqual(401, response.status_code)
        tus.cleanup()

    def test_tus_can_upload_pdf_file(self):
        # initialize the upload with POST
        pdf_file_path = os.path.join(os.path.dirname(__file__), UPLOAD_PDF_FILENAME)
        pdf_file_size = os.path.getsize(pdf_file_path)
        metadata = _prepare_metadata(UPLOAD_PDF_FILENAME, UPLOAD_PDF_MIMETYPE)
        response = self.api_session.post(
            self.upload_url,
            headers={
                "Tus-Resumable": "1.0.0",
                "Upload-Length": str(pdf_file_size),
                "Upload-Metadata": metadata,
            },
        )
        self.assertEqual(response.status_code, 201)
        location = response.headers["Location"]

        # upload the data with PATCH
        with open(pdf_file_path, "rb") as pdf_file:
            response = self.api_session.patch(
                location,
                headers={
                    "Content-Type": "application/offset+octet-stream",
                    "Upload-Offset": "0",
                    "Tus-Resumable": "1.0.0",
                },
                data=pdf_file,
            )
        self.assertEqual(response.status_code, 204)

        transaction.commit()
        self.assertEqual([UPLOAD_PDF_FILENAME], self.folder.contentIds())

    def test_tus_can_upload_text_file(self):
        # initialize the upload with POST
        metadata = _prepare_metadata(UPLOAD_FILENAME, UPLOAD_MIMETYPE)
        response = self.api_session.post(
            self.upload_url,
            headers={
                "Tus-Resumable": "1.0.0",
                "Upload-Length": str(UPLOAD_LENGTH),
                "Upload-Metadata": metadata,
            },
        )
        self.assertEqual(response.status_code, 201)
        location = response.headers["Location"]

        # upload the data with PATCH
        response = self.api_session.patch(
            location,
            headers={
                "Content-Type": "application/offset+octet-stream",
                "Upload-Offset": "0",
                "Tus-Resumable": "1.0.0",
            },
            data=BytesIO(UPLOAD_DATA),
        )
        self.assertEqual(response.status_code, 204)

    def test_tus_can_replace_pdf_file(self):
        # Create a test file
        self.file = api.content.create(
            container=self.portal, type="File", id="testfile", title="Testfile"
        )
        transaction.commit()
        # initialize the upload with POST
        pdf_file_path = os.path.join(os.path.dirname(__file__), UPLOAD_PDF_FILENAME)
        pdf_file_size = os.path.getsize(pdf_file_path)
        metadata = _prepare_metadata(UPLOAD_PDF_FILENAME, UPLOAD_PDF_MIMETYPE)
        response = self.api_session.post(
            f"{self.file.absolute_url()}/@tus-replace",
            headers={
                "Tus-Resumable": "1.0.0",
                "Upload-Length": str(pdf_file_size),
                "Upload-Metadata": metadata,
            },
        )
        self.assertEqual(response.status_code, 201)
        location = response.headers["Location"]

        # upload the data with PATCH
        with open(pdf_file_path, "rb") as pdf_file:
            response = self.api_session.patch(
                location,
                headers={
                    "Content-Type": "application/offset+octet-stream",
                    "Upload-Offset": "0",
                    "Tus-Resumable": "1.0.0",
                },
                data=pdf_file,
            )
        self.assertEqual(response.status_code, 204)

        transaction.commit()
        self.assertEqual(UPLOAD_PDF_FILENAME, self.file.file.filename)
        self.assertEqual(pdf_file_size, self.file.file.size)

    def test_create_with_tus_fires_proper_events(self):
        sm = getGlobalSiteManager()
        fired_events = []

        def record_event(event):
            fired_events.append(event.__class__.__name__)

        sm.registerHandler(record_event, (IObjectCreatedEvent,))
        sm.registerHandler(record_event, (IObjectWillBeAddedEvent,))
        sm.registerHandler(record_event, (IObjectAddedEvent,))
        sm.registerHandler(record_event, (IObjectModifiedEvent,))

        # initialize the upload with POST
        pdf_file_path = os.path.join(os.path.dirname(__file__), UPLOAD_PDF_FILENAME)
        pdf_file_size = os.path.getsize(pdf_file_path)
        metadata = _prepare_metadata(UPLOAD_PDF_FILENAME, UPLOAD_PDF_MIMETYPE)
        response = self.api_session.post(
            self.upload_url,
            headers={
                "Tus-Resumable": "1.0.0",
                "Upload-Length": str(pdf_file_size),
                "Upload-Metadata": metadata,
            },
        )
        self.assertEqual(response.status_code, 201)
        location = response.headers["Location"]

        # upload the data with PATCH
        with open(pdf_file_path, "rb") as pdf_file:
            response = self.api_session.patch(
                location,
                headers={
                    "Content-Type": "application/offset+octet-stream",
                    "Upload-Offset": "0",
                    "Tus-Resumable": "1.0.0",
                },
                data=pdf_file,
            )
        self.assertEqual(response.status_code, 204)

        self.assertEqual(
            fired_events,
            [
                "ObjectCreatedEvent",
                "ObjectWillBeAddedEvent",
                "ObjectAddedEvent",
                "ContainerModifiedEvent",
            ],
        )

        sm.unregisterHandler(record_event, (IObjectCreatedEvent,))
        sm.unregisterHandler(record_event, (IObjectWillBeAddedEvent,))
        sm.unregisterHandler(record_event, (IObjectAddedEvent,))
        sm.unregisterHandler(record_event, (IObjectModifiedEvent,))

    def test_replace_with_tus_fires_proper_events(self):
        # Create a test file
        self.file = api.content.create(
            container=self.portal, type="File", id="testfile", title="Testfile"
        )
        transaction.commit()

        sm = getGlobalSiteManager()
        fired_events = []

        def record_event(event):
            fired_events.append(event.__class__.__name__)

        sm.registerHandler(record_event, (IObjectCreatedEvent,))
        sm.registerHandler(record_event, (IObjectWillBeAddedEvent,))
        sm.registerHandler(record_event, (IObjectAddedEvent,))
        sm.registerHandler(record_event, (IObjectModifiedEvent,))

        # initialize the upload with POST
        pdf_file_path = os.path.join(os.path.dirname(__file__), UPLOAD_PDF_FILENAME)
        pdf_file_size = os.path.getsize(pdf_file_path)
        metadata = _prepare_metadata(UPLOAD_PDF_FILENAME, UPLOAD_PDF_MIMETYPE)
        response = self.api_session.post(
            f"{self.file.absolute_url()}/@tus-replace",
            headers={
                "Tus-Resumable": "1.0.0",
                "Upload-Length": str(pdf_file_size),
                "Upload-Metadata": metadata,
            },
        )
        self.assertEqual(response.status_code, 201)
        location = response.headers["Location"]

        # upload the data with PATCH
        with open(pdf_file_path, "rb") as pdf_file:
            response = self.api_session.patch(
                location,
                headers={
                    "Content-Type": "application/offset+octet-stream",
                    "Upload-Offset": "0",
                    "Tus-Resumable": "1.0.0",
                },
                data=pdf_file,
            )
        self.assertEqual(response.status_code, 204)

        self.assertEqual(fired_events, ["ObjectModifiedEvent"])

        sm.unregisterHandler(record_event, (IObjectCreatedEvent,))
        sm.unregisterHandler(record_event, (IObjectWillBeAddedEvent,))
        sm.unregisterHandler(record_event, (IObjectAddedEvent,))
        sm.unregisterHandler(record_event, (IObjectModifiedEvent,))

    def tearDown(self):
        self.api_session.close()
        client_home = os.environ.get("CLIENT_HOME")
        tmp_dir = os.path.join(client_home, "tus-uploads")
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)


class CORSTestPolicy(CORSPolicy):
    allow_origin = ["*"]
    allow_methods = ["DELETE", "GET", "OPTIONS", "PATCH", "POST", "PUT"]
    allow_credentials = True
    allow_headers = [
        "Accept",
        "Authorization",
        "Origin",
        "X-Requested-With",
        "Content-Type",
        "Tus-Resumable",
        "Upload-Length",
        "Upload-Offset",
    ]
    expose_header = []
    max_age = 3600


class TestTUSUploadWithCORS(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        provideAdapter(
            CORSTestPolicy, adapts=(Interface, IBrowserRequest), provides=ICORSPolicy
        )
        self.portal = self.layer["portal"]
        self.api_session = RelativeSession(self.portal.absolute_url(), test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        self.upload_url = f"{self.portal.absolute_url()}/@tus-upload"

    def test_cors_preflight_for_post_contains_tus_headers(self):
        response = self.api_session.options(
            self.upload_url,
            headers={
                "Origin": "http://myhost.net",
                "Access-Control-Request-Method": "POST",
                "Access-Control-Request-Headers": "Tus-Resumable,Upload-Length",
            },
        )
        self.assertIn("Tus-Resumable", response.headers)
        self.assertIn("Tus-Version", response.headers)
        self.assertIn("Tus-Extension", response.headers)

    def test_cors_preflight_for_patch_contains_tus_headers(self):
        response = self.api_session.options(
            self.upload_url,
            headers={
                "Origin": "http://myhost.net",
                "Access-Control-Request-Method": "PATCH",
                "Access-Control-Request-Headers": "Content-Type,Tus-Resumable,Upload-Offset",
            },
        )
        self.assertIn("Tus-Resumable", response.headers)
        self.assertIn("Tus-Version", response.headers)
        self.assertIn("Tus-Extension", response.headers)

    def test_cors_preflight_for_head_contains_tus_headers(self):
        response = self.api_session.options(
            self.upload_url,
            headers={
                "Origin": "http://myhost.net",
                "Access-Control-Request-Method": "HEAD",
                "Access-Control-Request-Headers": "Tus-Resumable",
            },
        )
        self.assertIn("Tus-Resumable", response.headers)
        self.assertIn("Tus-Version", response.headers)
        self.assertIn("Tus-Extension", response.headers)

    def tearDown(self):
        self.api_session.close()
        gsm = getGlobalSiteManager()
        gsm.unregisterAdapter(CORSTestPolicy, (Interface, IBrowserRequest), ICORSPolicy)


class TestTUSUpload(unittest.TestCase):
    def test_tmp_dir_gets_created_in_client_home(self):
        tus = TUSUpload("myuid")
        self.assertTrue(os.path.isdir(tus.tmp_dir))
        tus.cleanup()

    def test_use_tus_tmp_dir_if_provided(self):
        tus_upload_dir = tempfile.mkdtemp()
        os.environ["TUS_TMP_FILE_DIR"] = tus_upload_dir
        tus = TUSUpload("myuid")
        self.assertEqual(tus_upload_dir, tus.tmp_dir)
        tus.cleanup()

    def test_metadata_gets_stored_if_provided(self):
        tus = TUSUpload("myuid", {"length": 1024, "filename": "test.pdf"})
        self.assertIn("filename", tus.metadata())
        self.assertEqual("test.pdf", tus.metadata()["filename"])
        tus.cleanup()

    def test_length_returns_total_length_if_set(self):
        tus = TUSUpload("myuid", {"length": 1024})
        self.assertEqual(1024, tus.length())
        tus.cleanup()

    def test_length_returns_zero_if_not_set(self):
        tus = TUSUpload("myuid")
        self.assertEqual(0, tus.length())
        tus.cleanup()

    def test_offset_returns_zero_if_file_doesnt_exist(self):
        tus = TUSUpload("myuid", {"length": 1024})
        self.assertEqual(0, tus.offset())
        tus.cleanup()

    def test_offset_returns_size_of_current_file(self):
        tus = TUSUpload("myuid", {"length": 1024})
        tus.write(BytesIO(b"0123456789"))
        self.assertEqual(10, tus.offset())
        tus.cleanup()

    def test_write_creates_new_file(self):
        tus = TUSUpload("myuid", {"length": 1024})
        tus.write(BytesIO(b"0123456789"))
        self.assertTrue(os.path.isfile(tus.filepath))
        tus.cleanup()

    def test_write_appends_to_file_at_given_offset(self):
        tus = TUSUpload("myuid", {"length": 1024})
        tus.write(BytesIO(b"0123456789"))
        tus.write(BytesIO(b"abc"), 10)
        self.assertEqual(13, tus.offset())
        with open(tus.filepath, "rb") as f:
            data = f.read()
        self.assertEqual(b"0123456789abc", data)
        tus.cleanup()

    def test_write_sets_finished_flag(self):
        tus = TUSUpload("myuid", {"length": 10})
        tus.write(BytesIO(b"0123456789"))
        self.assertTrue(tus.finished)
        tus.cleanup()

    def test_metadata_returns_empty_dict_if_no_metadata_has_been_set(self):
        tus = TUSUpload("myuid")
        self.assertEqual({}, tus.metadata())
        tus.cleanup()

    def test_expires_returns_expiration_time_of_current_upload(self):
        tus = TUSUpload("myuid", {"length": 1024})
        tus.write(BytesIO(b"0123456789"))
        self.assertGreater(DateTime(tus.expires()), DateTime())
        tus.cleanup()

    def test_cleanup_removes_upload_file(self):
        tus = TUSUpload("myuid", {"length": 1024})
        tus.write(BytesIO(b"0123456789"))
        tus.cleanup()
        self.assertFalse(os.path.exists(tus.filepath))

    def test_cleanup_removes_metadata_file(self):
        tus = TUSUpload("myuid", {"length": 1024})
        tus.cleanup()
        self.assertFalse(os.path.exists(tus.metadata_path))

    def test_cleanup_expired_files(self):
        tus = TUSUpload("myuid")
        filepath = os.path.join(tus.tmp_dir, "tus_upload_12345")
        metadata_path = os.path.join(tus.tmp_dir, "tus_upload_12345.json")
        metadata_only_path = os.path.join(tus.tmp_dir, "tus_upload_67890.json")
        open(filepath, "wb").close()
        os.utime(filepath, (946684800.0, 946684800.0))
        open(metadata_path, "wb").close()
        os.utime(metadata_path, (946684800.0, 946684800.0))
        open(metadata_only_path, "wb").close()
        os.utime(metadata_only_path, (946684800.0, 946684800.0))
        tus.cleanup_expired()
        self.assertFalse(os.path.exists(filepath))
        self.assertFalse(os.path.exists(metadata_path))
        self.assertFalse(os.path.exists(metadata_only_path))
        tus.cleanup()

    def tearDown(self):
        client_home = os.environ.get("CLIENT_HOME")
        tmp_dir = os.path.join(client_home, "tus-uploads")
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)



from persistent.list import PersistentList
from plone.folder.default import DefaultOrdering
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from zope.annotation.interfaces import IAnnotatable
from zope.annotation.interfaces import IAnnotations


ORDER_KEY = DefaultOrdering.ORDER_KEY
QUERY = {"is_folderish": True}


def safe_utf8(to_utf8):
    if isinstance(to_utf8, str):
        to_utf8 = to_utf8.encode("utf-8")
    return to_utf8


def ensure_child_ordering_object_ids_are_native_strings(container):
    """Make sure the ordering stored on parent contains only native_string
    object ids.

    This function can be used to fix ordering object ids stored on a parent
    object in a `DefaultOrdering` ordering adapter. When changing object
    ordering via PATCH request we used to incorrectly store ids of reordered
    resouces as unicode instead of a bytestring (on python 2). This
    lead to mixed types being stored in the ordering annotations and
    subsequently mixed types being returned when calling `objectIds` of a
    container.

    The problem only exists with python 2 so we do nothing when we are
    called on python 3 by mistake.
    """
    return

    if not IAnnotatable.providedBy(container):
        return

    annotations = IAnnotations(container)
    if ORDER_KEY not in annotations:
        return

    fixed_ordering = PersistentList(
        safe_utf8(item_id) for item_id in annotations[ORDER_KEY]
    )
    annotations[ORDER_KEY] = fixed_ordering


class FixOrderingView(BrowserView):
    """Attempt to fix ordering for all potentially affected objects.

    By default will fix ordering object ids for every object that considers
    itself folderish.

    The problem only exists with python 2 so we do nothing when we are
    called on python 3 by mistake.
    """

    def __call__(self):
        return "Aborted, fixing ordering is only necessary on python 2."

        catalog = getToolByName(self.context, "portal_catalog")
        for brain in catalog.unrestrictedSearchResults(QUERY):
            folderish = brain.getObject()
            ensure_child_ordering_object_ids_are_native_strings(folderish)

        return "Done."



from plone.locking.interfaces import ILockable
from plone.locking.interfaces import INonStealableLock
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from plone.restapi.services.locking import lock_info
from zope.interface import alsoProvides
from zope.interface import noLongerProvides

import plone.protect.interfaces


class Lock(Service):
    """Unlock an object"""

    def reply(self):
        lockable = ILockable(self.context, None)
        if lockable is None:
            return lock_info(self.context)

        data = json_body(self.request)

        # Remove lock by the same user or steal it
        if lockable.can_safely_unlock() or data.get("force"):
            lockable.unlock()

            if INonStealableLock.providedBy(self.context):
                noLongerProvides(self.context, INonStealableLock)

            # Disable CSRF protection
            if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
                alsoProvides(
                    self.request, plone.protect.interfaces.IDisableCSRFProtection
                )

        return lock_info(self.context)



from plone.dexterity.interfaces import IDexterityFTI
from plone.restapi.indexers import SearchableText_blocks
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from uuid import uuid4
from zope.component import queryUtility

import unittest


DRAFTJS_BLOCK = {
    "@type": "text",
    "text": {
        "blocks": [
            {
                "data": {},
                "depth": 0,
                "entityRanges": [],
                "inlineStyleRanges": [],
                "key": "behki",
                "text": "Plone is a powerful content management system",
                "type": "unstyled",
            }
        ],
        "entityMap": {},
    },
}

SLATE_BLOCK = {
    "@type": "slate",
    "plaintext": "Follow Plone Conference",
    "value": [{"children": [{"text": "Follow Plone Conference"}]}],
}


TABLE_BLOCK = {
    "@type": "table",
    "table": {
        "rows": [
            {
                "key": "a",
                "cells": [
                    {
                        "type": "data",
                        "key": "b",
                        "value": {
                            "blocks": [
                                {
                                    "data": {},
                                    "depth": 0,
                                    "entityRanges": [],
                                    "inlineStyleRanges": [],
                                    "key": "fgm98",
                                    "text": "My header",
                                    "type": "header-two",
                                },
                            ],
                            "entityMap": {},
                        },
                    },
                ],
            },
            {
                "key": "c",
                "cells": [
                    {
                        "type": "data",
                        "key": "d",
                        "value": {
                            "blocks": [
                                {
                                    "data": {},
                                    "depth": 0,
                                    "entityRanges": [],
                                    "inlineStyleRanges": [],
                                    "key": "mmm99",
                                    "text": "My data",
                                    "type": "text",
                                },
                            ],
                            "entityMap": {},
                        },
                    },
                ],
            },
        ]
    },
}


class TestSearchableTextIndexer(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        fti = queryUtility(IDexterityFTI, name="Document")
        behavior_list = [a for a in fti.behaviors]
        behavior_list.append("volto.blocks")
        fti.behaviors = tuple(behavior_list)
        self.portal.invokeFactory(
            "Document",
            id="doc1",
            title="Title is here",
            description="Description is there",
            blocks={},
            blocks_layout={"items": []},
        )
        self.document = self.portal["doc1"]

    @staticmethod
    def _extract_searchable_text(obj):
        indexer = SearchableText_blocks(obj)
        return indexer()

    @staticmethod
    def _add_blocks(obj, raw_blocks):
        blocks = {str(uuid4()): block for block in raw_blocks}
        layout = list(blocks.keys())
        obj.blocks = blocks
        obj.blocks_layout["items"] = layout

    def test_indexer_no_blocks(self):
        result = self._extract_searchable_text(self.document)
        self.assertIn("Title is here", result)
        self.assertIn("Description is there", result)

    def test_indexer_block_has_searchableText(self):
        document = self.document
        self._add_blocks(
            document,
            [{"@type": "new-block", "attribute": "bar", "searchableText": "Foo Bar"}],
        )
        result = self._extract_searchable_text(document)
        self.assertIn("Title is here Description is there", result)
        self.assertIn("Foo Bar", result)

    def test_indexer_multiple_blocks(self):
        document = self.document
        self._add_blocks(
            document,
            [
                {"@type": "a-block", "key": "value", "searchableText": "Plone is a"},
                {"@type": "a-block", "attribute": "bar", "searchableText": "CMS"},
            ],
        )
        result = self._extract_searchable_text(document)
        self.assertIn("Plone is a CMS ", result)

    def test_indexer_block_slate_bbb(self):
        document = self.document
        self._add_blocks(
            document,
            [
                SLATE_BLOCK,
            ],
        )
        result = self._extract_searchable_text(document)
        self.assertIn("Title is here Description is there", result)
        self.assertIn("Follow Plone Conference", result)

    def test_indexer_block_text(self):
        document = self.document
        self._add_blocks(
            document,
            [
                DRAFTJS_BLOCK,
            ],
        )
        result = self._extract_searchable_text(document)
        self.assertIn("Title is here Description is there", result)
        self.assertIn("Plone is a powerful content management system", result)

    def test_indexer_block_table(self):
        document = self.document
        self._add_blocks(
            document,
            [
                TABLE_BLOCK,
            ],
        )
        result = self._extract_searchable_text(document)
        self.assertIn("Title is here Description is there", result)
        self.assertIn("My data", result)

    def test_indexer_block_with_subblocks(self):
        document = self.document
        block = {
            "@type": "complex-block",
            "blocks": {
                "ad625c27-670a-412e-821d-57870b6e82f1": SLATE_BLOCK,
                "80c5d58d-4baf-4adc-8ef4-36810df5628d": DRAFTJS_BLOCK,
                "4061b21d-c971-4ccc-a719-d8b24214fd8b": {
                    "@type": "old-complex-block",
                    "data": {
                        "blocks": {
                            "116feae0-7aa3-40f4-a776-fc893d78b353": TABLE_BLOCK,
                        }
                    },
                },
            },
        }
        self._add_blocks(
            document,
            [
                block,
            ],
        )
        result = self._extract_searchable_text(document)
        # From Slate sub block
        self.assertIn("Follow Plone Conference", result)
        # From DraftJS
        self.assertIn("Plone is a powerful content management system", result)
        # From Table block
        self.assertIn("My data", result)



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import unittest


class TestControlpanelsEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def tearDown(self):
        self.api_session.close()

    def test_get_listing(self):
        # Do we get a list with at least one item?
        response = self.api_session.get("/@controlpanels")
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertIs(type(data), list)
        self.assertGreater(len(data), 0)

    def test_get_item_nonexisting(self):
        response = self.api_session.get("/@controlpanels/no-way-jose")
        self.assertEqual(404, response.status_code)

    def test_get_item(self):
        response = self.api_session.get("/@controlpanels/editing")
        self.assertEqual(200, response.status_code)

    def test_all_controlpanels(self):
        # make sure all define controlpanels serialize
        response = self.api_session.get("/@controlpanels")
        for item in response.json():
            response = self.api_session.get(item["@id"])
            self.assertEqual(
                200,
                response.status_code,
                "{} failed: {}".format(item["@id"], response.json()),
            )

    def test_patch_needs_parameter(self):
        response = self.api_session.patch("/@controlpanels")
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            "Missing parameter controlpanelname", response.json()["message"]
        )

    def test_update(self):
        # get current settings, switch them and check if it changed
        response = self.api_session.get("/@controlpanels/editing")
        old_data = response.json()["data"]

        # switch values and set
        new_values = {
            "ext_editor": not old_data["ext_editor"],
            "lock_on_ttw_edit": not old_data["lock_on_ttw_edit"],
        }
        response = self.api_session.patch("/@controlpanels/editing", json=new_values)

        # check if the values changed
        response = self.api_session.get("/@controlpanels/editing")
        self.assertNotEqual(response.json(), old_data)

    def test_update_all(self):
        # Mail is in faulty state by default
        self.api_session.patch(
            "/@controlpanels/mail",
            json={
                "email_from_address": "admin@local.local",
                "email_from_name": "Jos Henken",
            },
        )

        # make sure all define controlpanels deserialize
        response = self.api_session.get("/@controlpanels")
        for item in response.json():
            # get current data
            response = self.api_session.get(item["@id"])

            # store the outputted data
            response = self.api_session.patch(item["@id"], json=response.json()["data"])
            self.assertEqual(
                204,
                response.status_code,
                "{} failed: {}".format(item["@id"], response.content),
            )

    def test_update_required(self):
        KEY = "email_charset"
        URL = "/@controlpanels/mail"
        # sanity check
        response = self.api_session.get(URL)
        response = response.json()
        self.assertIn(KEY, response["schema"]["required"])

        response = self.api_session.patch(URL, json={KEY: None})

        self.assertEqual(response.status_code, 400)
        response = response.json()
        self.assertIn("message", response)
        self.assertIn("Required input is missing.", response["message"])



from plone.keyring.interfaces import IKeyManager
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from Products.CMFCore.utils import getToolByName
from zope.component import getUtility

import unittest


class TestJWTAuthenticationPlugin(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):

        self.portal = self.layer["portal"]
        uf = getToolByName(self.portal, "acl_users")
        self.plugin = uf["jwt_auth"]

    def test_challenge(self):
        request = self.layer["request"]
        response = request.response
        self.plugin.challenge(request, request.response)
        self.assertEqual(401, response.getStatus())
        self.assertEqual('Bearer realm="Zope"', response.getHeader("WWW-Authenticate"))

    def test_extract_credentials_without_authorization_header(self):
        request = self.layer["request"]
        request._auth = ""
        self.assertEqual(None, self.plugin.extractCredentials(request))

    def test_extract_credentials_with_other_authorization_header(self):
        request = self.layer["request"]
        request._auth = "Basic YWRtaW46YWRtaW4="
        self.assertEqual(None, self.plugin.extractCredentials(request))

    def test_extract_credentials_with_bearer_authorization_header(self):
        request = self.layer["request"]
        request._auth = (
            "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJhZG1pbiJ9."
            "PGnRccPTXeaxA8nzfytWewWRkizJa_ihI_3H6ec-Zbw"
        )
        self.assertEqual(
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJhZG1pbiJ9.PGnRccP"
            "TXeaxA8nzfytWewWRkizJa_ihI_3H6ec-Zbw",
            self.plugin.extractCredentials(request)["token"],
        )

    def test_authenticate_credentials_from_unknown_extractor(self):
        creds = {}
        creds["extractor"] = "credentials_basic_auth"
        self.assertEqual(None, self.plugin.authenticateCredentials(creds))

    def test_authenticate_credentials_with_invalid_token(self):
        creds = {}
        creds["extractor"] = "jwt_auth"
        creds["token"] = "invalid"
        self.assertEqual(None, self.plugin.authenticateCredentials(creds))

    def test_authenticate_credentials_without_subject(self):
        creds = {}
        creds["extractor"] = "jwt_auth"
        creds["token"] = (
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.e30.t-IDcSemACt8x4iTMCda8Yhe"
            "3iZaWbvV5XKSTbuAn0M"
        )
        self.assertEqual(None, self.plugin.authenticateCredentials(creds))

    def test_authenticate_credentials_with_valid_token(self):
        creds = {}
        creds["extractor"] = "jwt_auth"
        creds["token"] = self.plugin.create_token("admin")
        self.assertEqual(("admin", "admin"), self.plugin.authenticateCredentials(creds))

    def test_authenticate_credentials_returns_native_string(self):
        creds = {}
        creds["extractor"] = "jwt_auth"
        creds["token"] = self.plugin.create_token("admin")
        self.assertIsInstance(self.plugin.authenticateCredentials(creds)[0], str)

    def test_decode_token_after_key_rotation(self):
        token = self.plugin.create_token("admin", timeout=0)
        key_manager = getUtility(IKeyManager)
        key_manager.rotate()
        self.assertEqual({"sub": "admin"}, self.plugin._decode_token(token))

    def test_decode_with_static_secret(self):
        self.plugin.use_keyring = False
        token = self.plugin.create_token("admin", timeout=0)
        self.assertEqual({"sub": "admin"}, self.plugin._decode_token(token))

    def test_authenticate_credentials_with_stored_token(self):
        self.plugin.store_tokens = True
        creds = {}
        creds["extractor"] = "jwt_auth"
        creds["token"] = self.plugin.create_token("admin")
        self.assertEqual(("admin", "admin"), self.plugin.authenticateCredentials(creds))

    def test_authenticate_credentials_with_deleted_token_fails(self):
        self.plugin.store_tokens = True
        creds = {}
        creds["extractor"] = "jwt_auth"
        creds["token"] = self.plugin.create_token("admin")
        self.plugin.delete_token(creds["token"])
        self.assertEqual(None, self.plugin.authenticateCredentials(creds))



from zope.interface import implementer
from zope.publisher.interfaces import NotFound
from plone.restapi.controlpanels.interfaces import IControlpanel
from Products.CMFCore.utils import getToolByName


@implementer(IControlpanel)
class RegistryConfigletPanel:
    configlet = None
    configlet_id = None
    configlet_category_id = None
    schema = None

    schema_prefix = "plone"

    def _get_configlet(self):
        configlet_data = self.portal_cp.enumConfiglets(self.configlet_category_id)
        for action in configlet_data:
            if action["id"] == self.configlet_id:
                return action

    def _get_group_title(self):
        groups = [
            g
            for g in self.portal_cp.getGroups()
            if g["id"] == self.configlet["category"]
        ]
        return [g["title"] for g in groups][0]

    def __init__(self, context, request):
        self.context = context
        self.request = request

        self.portal_cp = getToolByName(self.context, "portal_controlpanel")

        self.configlet = self._get_configlet()
        if self.configlet:
            self.title = self.configlet["title"]
            self.group = self._get_group_title()

    def add(self, names):
        raise NotFound(self.context, names, self.request)

    def get(self, names):
        raise NotFound(self.context, names, self.request)

    def update(self, names):
        raise NotFound(self.context, names, self.request)

    def delete(self, names):
        raise NotFound(self.context, names, self.request)



from AccessControl import getSecurityManager
from Acquisition import aq_inner
from Acquisition import aq_parent
from plone.app.dexterity.behaviors.nextprevious import NextPreviousBase
from plone.registry.interfaces import IRegistry
from zope.component import getUtility


class NextPreviousFixed(NextPreviousBase):
    """
    Based on plone.app.dexterity.behaviors.nextprevious.NextPreviousBase
    but works for IPloneSite object
    """

    def __init__(self, context):
        self.context = context
        registry = getUtility(IRegistry)
        self.vat = registry.get("plone.types_use_view_action_in_listings", [])
        self.security = getSecurityManager()
        self.order = self.context.objectIds()


class NextPrevious:
    """Facade with more pythonic interface"""

    def __init__(self, context):
        self.context = context
        self.parent = aq_parent(aq_inner(context))
        self.nextprev = NextPreviousFixed(self.parent)

    @property
    def next(self):
        """return info about the next item in the container"""
        if getattr(self.parent, "_ordering", "") == "unordered":
            # Unordered folder
            return {}
        data = self.nextprev.getNextItem(self.context)
        if data is None:
            return {}
        return {
            "@id": data["url"].lstrip("/view"),
            "@type": data["portal_type"],
            "title": data["title"],
            "description": data["description"],
        }

    @property
    def previous(self):
        """return info about the previous item in the container"""
        if getattr(self.parent, "_ordering", "") == "unordered":
            # Unordered folder
            return {}
        data = self.nextprev.getPreviousItem(self.context)
        if data is None:
            return {}
        return {
            "@id": data["url"].lstrip("/view"),
            "@type": data["portal_type"],
            "title": data["title"],
            "description": data["description"],
        }
